/* src/vm/jit/arm/codegen.c - machine code generator for arm

   Copyright (C) 1996-2005 R. Grafl, A. Krall, C. Kruegel, C. Oates,
   R. Obermaisser, M. Platter, M. Probst, S. Ring, E. Steiner,
   C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich, J. Wenninger,
   Institut f. Computersprachen - TU Wien

   This file is part of CACAO.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   Contact: cacao@complang.tuwien.ac.at

   Authors: Michael Starzinger

   $Id: codegen.c,v 1.22 2005/07/12 16:07:06 michi Exp $

*/


#include <stdio.h>

#include "config.h"

#include "md-abi.h"
#include "md-abi.inc"

#include "vm/jit/arm/arch.h"
#include "vm/jit/arm/codegen.h"
#include "vm/jit/arm/types.h"
#include "vm/jit/arm/asmoffsets.h"

#include "cacao/cacao.h"
#include "native/native.h"
#include "vm/builtin.h"
#include "vm/global.h"
#include "vm/loader.h"
#include "vm/stringlocal.h"
#include "vm/tables.h"
#include "vm/jit/asmpart.h"
#include "vm/jit/codegen.inc"
#include "vm/jit/jit.h"

#if defined(LSRA)
# include "vm/jit/lsra.h"
# include "vm/jit/lsra.inc"
#endif

#include "vm/jit/parse.h"
#include "vm/jit/patcher.h"
#include "vm/jit/reg.h"
#include "vm/jit/reg.inc"


/* function codegen ************************************************************

   generates machine code

*******************************************************************************/

void codegen(methodinfo *m, codegendata *cd, registerdata *rd)
{
	s4              i, t, len;
	s4              s1, s2, s3, d;
	/*ptrint*/s4          a;
	s4             *mcodeptr;
	stackptr        src;
	varinfo        *var;
	basicblock     *bptr;
	instruction    *iptr;
	exceptiontable *ex;

	s4              parentargs_base;
	s4              spilledregs_num;
	s4              savedregs_num;
	u2              savedregs_bitmask;
	u2              currentline;

	methodinfo         *lm;             /* local methodinfo for ICMD_INVOKE* */
	builtintable_entry *bte;

	methoddesc         *md;

	/* space to save used callee saved registers */
	savedregs_num = (m->isleafmethod) ? 0 : 1;        /* space to save the LR */
	savedregs_num += (INT_SAV_CNT - rd->savintreguse);
	savedregs_num += (FLT_SAV_CNT - rd->savfltreguse);
	spilledregs_num = rd->memuse;
	parentargs_base = spilledregs_num + savedregs_num;

	/* create method header */
	(void) dseg_addaddress(cd, m);                          /* MethodPointer  */
	(void) dseg_adds4(cd, parentargs_base * 4);             /* FrameSize      */
#if defined(USE_THREADS)
	/* IsSync contains the offset relative to the stack pointer for the
	   argument of monitor_exit used in the exception handler. Since the
	   offset could be zero and give a wrong meaning of the flag it is
	   offset by one.
	*/
	if (checksync && (m->flags & ACC_SYNCHRONIZED))
		(void) dseg_adds4(cd, (rd->memuse + 1) * 4);                /* IsSync */
	else
#endif
	(void) dseg_adds4(cd, 0);                                       /* IsSync */
	(void) dseg_adds4(cd, m->isleafmethod);                         /* IsLeaf */
	(void) dseg_adds4(cd, INT_SAV_CNT - rd->savintreguse);         /* IntSave */
	(void) dseg_adds4(cd, FLT_SAV_CNT - rd->savfltreguse);         /* FltSave */
	(void) dseg_addlinenumbertablesize(cd);
	(void) dseg_adds4(cd, cd->exceptiontablelength);           /* ExTableSize */

	/* create exception table */
	for (ex = cd->exceptiontable; ex != NULL; ex = ex->down) {
		dseg_addtarget(cd, ex->start);
		dseg_addtarget(cd, ex->end);
		dseg_addtarget(cd, ex->handler);
		(void) dseg_addaddress(cd, ex->catchtype.cls);
	}

	/* initialize mcode variables */
	mcodeptr = (s4*) cd->mcodebase;
	cd->mcodeend = (s4*) (cd->mcodebase + cd->mcodesize);
	/* TODO: mcodecheck*/ MCODECHECK(128 + m->paramcount);

	/* save return address and used callee saved registers */
	savedregs_bitmask = 0;
	if (!m->isleafmethod)
		savedregs_bitmask = (1<<REG_LR);
	for (i = INT_SAV_CNT - 1; i >= rd->savintreguse; i--)
		savedregs_bitmask |= (1<<(rd->savintregs[i]));
	for (i = FLT_SAV_CNT - 1; i >= rd->savfltreguse; i--) {
		log_text("!!! CODEGEN: floating-point callee saved registers are not saved to stack (SEVERE! STACK IS MESSED UP!)");
		/* TODO: floating-point */
	}
	if (savedregs_bitmask) {
		M_STMFD(savedregs_bitmask, REG_SP);
	}

	/* create additional stack frame for spilled variables (if necessary) */
	if (spilledregs_num) {
		/* TODO: do i need this check? ask c_u! */
		if (spilledregs_num > 63)
			throw_cacao_exception_exit(string_java_lang_InternalError,
			                           "Too much spilled variables: %d (>63)", spilledregs_num);
		M_SUB_IMM(REG_SP, REG_SP, spilledregs_num * 4);
	}

	/* take arguments out of register or stack frame */
	/* TODO: make this work! */
	md = m->parseddesc;
	for (i = 0, len = 0; i < md->paramcount; i++) {
		s1 = md->params[i].regoff;
		t = md->paramtypes[i].type;
		var = &(rd->locals[len][t]);
		len += (IS_2_WORD_TYPE(t)) ? 2 : 1;          /* 2 word type arguments */
		if (var->type < 0)                           /* argument is not used  */
			continue;
		/* ATTENTION: we use interger registers for all arguments (even float) */
		if (IS_INT_LNG_TYPE(t)) {                    /* integer args          */
			if (!md->params[i].inmemory) {           /* register arguments    */
				s2 = ARGUMENT_REGS(t, s1);           /* get argument register */
				if (!(var->flags & INMEMORY)) {      /* reg arg -> register   */
					M_INTMOVE_TYPED(t, s2, var->regoff);
				} else {                             /* reg arg -> spilled    */
					if (IS_2_WORD_TYPE(t)) {
						M_STR(GET_LOW_REG(s2), REG_SP, 4 * (var->regoff+0));
						M_STR(GET_HIGH_REG(s2), REG_SP, 4 * (var->regoff+1));
					} else {
						M_STR(s2, REG_SP, 4 * var->regoff);
					}
				}
			} else {                                 /* stack arguments       */
				if (!(var->flags & INMEMORY)) {      /* stack arg -> register */
					if (IS_2_WORD_TYPE(t)) {
						M_LDR(GET_LOW_REG(var->regoff), REG_SP, 4 * (parentargs_base + s1 + 0));
						M_LDR(GET_HIGH_REG(var->regoff), REG_SP, 4 * (parentargs_base + s1 + 1));
					} else {
						M_LDR(var->regoff, REG_SP, 4 * (parentargs_base + s1));
					}
				} else {                             /* stack arg -> spilled  */
					/* Reuse Memory Position on Caller Stack */
					var->regoff = parentargs_base + s1;
				}
			}
		} else {                                     /* floating args         */
			if (!md->params[i].inmemory) {           /* register arguments    */
				s2 = ARGUMENT_REGS(t, s1);           /* get argument register */
				if (!(var->flags & INMEMORY)) {      /* reg arg -> register   */
					M_FLTLOAD_TYPED(t, s2, var->regoff);
				} else {                             /* reg arg -> spilled    */
					if (IS_2_WORD_TYPE(t)) {
						M_STR(GET_LOW_REG(s2), REG_SP, 4 * (var->regoff+0));
						M_STR(GET_HIGH_REG(s2), REG_SP, 4 * (var->regoff+1));
					} else {
						M_STR(s2, REG_SP, 4 * var->regoff);
					}
				}
			} else {                                 /* stack arguments       */
				if (!(var->flags & INMEMORY)) {      /* stack arg -> register */
					if (IS_2_WORD_TYPE(t))
						M_LDFD(var->regoff, REG_SP, 4 * (parentargs_base + s1));
					else
						M_LDFS(var->regoff, REG_SP, 4 * (parentargs_base + s1));
				} else {                             /* stack arg -> spilled  */
					/* Reuse Memory Position on Caller Stack */
					var->regoff = parentargs_base + s1;
				}
			}
			dolog("!!! ARGUMENTS: float and double not tested yet (%s.%s)", m->class->name->text, m->name->text);
		}
	}

	/* call monitorenter function */
#if defined(USE_THREADS)
	if (checksync && (m->flags & ACC_SYNCHRONIZED)) {
		log_text("!!! MONITOR_ENTER: not implemented yet");
	}
#endif

	/* call trace function */
	if (runverbose) {
		/* save argument registers to stack (including LR and IP) */
		/*M_SUB_IMM(REG_SP, REG_SP, (INT_ARG_CNT + FLT_ARG_CNT + 2) * 4);*/
		M_STMFD(0x0f | (1<<REG_LR) | (1<<REG_IP), REG_SP);

		/* prepare args for tracer */
		M_SUB_IMM(REG_SP, REG_SP, (2+2+1)*4); /* space for a3, a4 and m */
		ICONST(REG_ITMP1, 0);
		M_STR(2, REG_SP, 0);
		M_STR(REG_ITMP1, REG_SP, 4);
		M_STR(3, REG_SP, 8);
		M_STR(REG_ITMP1, REG_SP, 12);
		M_MOV(2, 1);
		M_MOV(1, REG_ITMP1);
		M_MOV(3, REG_ITMP1);
		a = dseg_addaddress(cd, (void*) m);
		M_LDR(REG_ITMP1, REG_IP, a);
		M_STR(REG_ITMP1, REG_SP, 16);

		/* call tracer here (we use a long branch) */
		M_LONGBRANCH(builtin_trace_args);

		M_ADD_IMM(REG_SP, REG_SP, (2+2+1)*4); /* free stack */

		/* restore argument registers from stack */
		M_LDMFD(0x0f | (1<<REG_LR) | (1<<REG_IP), REG_SP);
	}

	/* end of header generation */

	/* for all basic blocks */
	for (bptr = m->basicblocks; bptr != NULL; bptr = bptr->next)
	{
		bptr->mpc = (int) ((u1*) mcodeptr - cd->mcodebase);

		/* is this basic block reached? */
		if (bptr->flags < BBREACHED)
			continue;

		/* branch resolving */
		{
		branchref *brefs;
		for (brefs = bptr->branchrefs; brefs != NULL; brefs = brefs->next) {
			gen_resolvebranch((u1*) cd->mcodebase + brefs->branchpos, 
			                  brefs->branchpos, bptr->mpc);
			}
		}

		/* copy interface registers to their destination */
		src = bptr->instack;
		len = bptr->indepth;
		MCODECHECK(64+len);
		while (src != NULL) {
			/*log_text("!!! INSTACK: we have some regs here");*/
			len--;
			if ((len == 0) && (bptr->type != BBTYPE_STD)) {
				d = reg_of_var(rd, src, REG_ITMP1);
				M_INTMOVE(REG_ITMP1, d);
				store_reg_to_var_int(src, d);
			} else {
				d = reg_of_var(rd, src, REG_ITMP1);
				if ((src->varkind != STACKVAR)) {
					t = src->type;
					if (IS_FLT_DBL_TYPE(t)) {
						log_text("!!! INSTACK: flaot and double interface registers not implemented yet");
					} else {
						if (!(rd->interfaces[len][t].flags & INMEMORY)) {
							s1 = rd->interfaces[len][t].regoff;
							M_INTMOVE_TYPED(t,s1,d);
						} else {
							if (IS_2_WORD_TYPE(t)) {
								M_LDR(GET_LOW_REG(d), REG_SP, 4 * (rd->interfaces[len][t].regoff+0));
								M_LDR(GET_HIGH_REG(d), REG_SP, 4 * (rd->interfaces[len][t].regoff+1));
							} else {
								M_LDR(d, REG_SP, 4 * rd->interfaces[len][t].regoff);
							}
						}
						store_reg_to_var_int(src, d);
					}
				}
			}
			src = src->prev;
		}

		/* for all instructions */
		src = bptr->instack;
		len = bptr->icount;
		for (iptr = bptr->iinstr; len > 0; src = iptr->dst, len--, iptr++) {

			/* add line number */
			if (iptr->line != currentline) {
				dseg_addlinenumber(cd, iptr->line, (u1*)mcodeptr);
				currentline = iptr->line;
			}

			/* TODO: other checks as well!!! */ MCODECHECK(64);       /* an instruction usually needs < 64 words      */

			/* the big switch */
			switch (iptr->opc) {
			case ICMD_NOP:        /* ... ==> ...                                  */
				break;

			case ICMD_CHECKNULL:  /* ..., objectref  ==> ..., objectref           */
				load_var_to_reg_int(s1, src, REG_ITMP1);
				M_TST(s1, s1);
				M_BEQ(-2);
				codegen_addxnullrefs(cd, mcodeptr);
				break;

		/* constant operations ************************************************/
			case ICMD_ICONST:     /* ...  ==> ..., constant                       */
			                      /* op1 = 0, val.i = constant                    */
				d = reg_of_var(rd, iptr->dst, REG_ITMP1);
				ICONST(d, iptr->val.i);
				store_reg_to_var_int(iptr->dst, d);
				break;
			case ICMD_ACONST:     /* ... ==> ..., constant                        */
			                      /* op1 = 0, val.a = constant                    */
				d = reg_of_var(rd, iptr->dst, REG_ITMP1);
				ICONST(d, (u4) iptr->val.a);
				store_reg_to_var_int(iptr->dst, d);
				break;
			case ICMD_LCONST:     /* ...  ==> ..., constant                       */
			                      /* op1 = 0, val.l = constant                    */
				d = reg_of_var(rd, iptr->dst, PACK_REGS(REG_ITMP1, REG_ITMP2));
				LCONST(d, iptr->val.l);
				/*M_BREAKPOINT(0x1111);*/
				store_reg_to_var_int(iptr->dst, d);
				break;
			case ICMD_FCONST:     /* ...  ==> ..., constant                       */
			                      /* op1 = 0, val.f = constant                    */
				d = reg_of_var(rd, iptr->dst, REG_ITMP1);
				a = dseg_addfloat(cd, iptr->val.f);
				M_LDFS(d, REG_IP, a);
				store_reg_to_var_flt(iptr->dst, d);
				break;
			case ICMD_DCONST:     /* ...  ==> ..., constant                       */
			                      /* op1 = 0, val.d = constant                    */
				d = reg_of_var(rd, iptr->dst, REG_FTMP1);
				a = dseg_adddouble(cd, iptr->val.d);
				M_LDFD(d, REG_IP, a);
				/*M_BREAKPOINT(0xf1041);*/
				store_reg_to_var_flt(iptr->dst, d);
				break;

		/* load/store operations **********************************************/
			case ICMD_ILOAD:      /* ...  ==> ..., content of local variable      */
			case ICMD_ALOAD:      /* op1 = local variable                         */
			case ICMD_LLOAD:
				var = &(rd->locals[iptr->op1][iptr->opc - ICMD_ILOAD]);
				if (IS_2_WORD_TYPE(var->type))
					d = reg_of_var(rd, iptr->dst, PACK_REGS(REG_ITMP1, REG_ITMP2));
				else
					d = reg_of_var(rd, iptr->dst, REG_ITMP1);
				if ((iptr->dst->varkind == LOCALVAR) && (iptr->dst->varnum == iptr->op1))
					break;
				if (var->flags & INMEMORY) {
					if (IS_2_WORD_TYPE(var->type)) {
						M_LDR(GET_LOW_REG(d), REG_SP, 4 * (var->regoff+0));
						M_LDR(GET_HIGH_REG(d), REG_SP, 4 * (var->regoff+1));
					} else {
						M_LDR(d, REG_SP, 4 * var->regoff);
					}
				} else {
					M_INTMOVE_TYPED(var->type, var->regoff, d);
				}
				store_reg_to_var_int(iptr->dst, d);
				break;

			case ICMD_FLOAD:      /* ...  ==> ..., content of local variable      */
			case ICMD_DLOAD:      /* op1 = local variable                         */
				d = reg_of_var(rd, iptr->dst, REG_FTMP1);
				if ((iptr->dst->varkind == LOCALVAR) && (iptr->dst->varnum == iptr->op1))
					break;
				var = &(rd->locals[iptr->op1][iptr->opc - ICMD_ILOAD]);
				if (var->flags & INMEMORY) {
					if (var->type == TYPE_DBL)
						M_LDFD(d, REG_SP, 4 * var->regoff);
					else
						M_LDFS(d, REG_SP, 4 * var->regoff);
				} else {
					M_FLTMOVE_TYPED(var->type, var->regoff, d);
				}
				store_reg_to_var_flt(iptr->dst, d);
				break;

			case ICMD_ISTORE:     /* ..., value  ==> ...                          */
			case ICMD_ASTORE:     /* op1 = local variable                         */
			case ICMD_LSTORE:
				if ((src->varkind == LOCALVAR) && (src->varnum == iptr->op1))
					break;
				var = &(rd->locals[iptr->op1][iptr->opc - ICMD_ISTORE]);
				if (var->flags & INMEMORY) {
					if (IS_2_WORD_TYPE(var->type)) {
						load_var_to_reg_int(s1, src, PACK_REGS(REG_ITMP1, REG_ITMP2));
						M_STR(GET_LOW_REG(s1), REG_SP, 4 * (var->regoff+0));
						M_STR(GET_HIGH_REG(s1), REG_SP, 4 * (var->regoff+1));
					} else {
						load_var_to_reg_int(s1, src, REG_ITMP1);
						M_STR(s1, REG_SP, 4 * var->regoff);
					}
				} else {
					load_var_to_reg_int(s1, src, var->regoff);
					M_INTMOVE_TYPED(var->type, s1, var->regoff);
				}
				break;

			case ICMD_FSTORE:     /* ..., value  ==> ...                          */
			case ICMD_DSTORE:     /* op1 = local variable                         */
				if ((src->varkind == LOCALVAR) && (src->varnum == iptr->op1))
					break;
				var = &(rd->locals[iptr->op1][iptr->opc - ICMD_ISTORE]);
				if (var->flags & INMEMORY) {
					load_var_to_reg_flt(s1, src, REG_FTMP1);
					if (var->type == TYPE_DBL)
						M_STFD(s1, REG_SP, 4 * var->regoff);
					else
						M_STFS(s1, REG_SP, 4 * var->regoff);
				} else {
					load_var_to_reg_flt(s1, src, var->regoff);
					M_FLTMOVE_TYPED(var->type, s1, var->regoff);
				}
				break;

		/* pop/dup/swap operations ********************************************/
		/* attention: double and longs are only one entry in CACAO ICMDs      */
			case ICMD_POP:        /* ..., value  ==> ...                          */
			case ICMD_POP2:       /* ..., value, value  ==> ...                   */
				break;

			case ICMD_DUP:        /* ..., a ==> ..., a, a                         */
				M_COPY(src, iptr->dst);
				break;
			case ICMD_DUP_X1:     /* ..., a, b ==> ..., b, a, b                   */
				M_COPY(src,       iptr->dst);
				M_COPY(src->prev, iptr->dst->prev);
				M_COPY(iptr->dst, iptr->dst->prev->prev);
				break;
			case ICMD_DUP_X2:     /* ..., a, b, c ==> ..., c, a, b, c             */
				M_COPY(src,             iptr->dst);
				M_COPY(src->prev,       iptr->dst->prev);
				M_COPY(src->prev->prev, iptr->dst->prev->prev);
				M_COPY(iptr->dst,       iptr->dst->prev->prev->prev);
				break;

			case ICMD_DUP2:       /* ..., a, b ==> ..., a, b, a, b                */
				M_COPY(src,       iptr->dst);
				M_COPY(src->prev, iptr->dst->prev);
				break;
			case ICMD_DUP2_X1:    /* ..., a, b, c ==> ..., b, c, a, b, c          */
				M_COPY(src,             iptr->dst);
				M_COPY(src->prev,       iptr->dst->prev);
				M_COPY(src->prev->prev, iptr->dst->prev->prev);
				M_COPY(iptr->dst,       iptr->dst->prev->prev->prev);
				M_COPY(iptr->dst->prev, iptr->dst->prev->prev->prev->prev);
				break;
			case ICMD_DUP2_X2:    /* ..., a, b, c, d ==> ..., c, d, a, b, c, d    */
				M_COPY(src,                   iptr->dst);
				M_COPY(src->prev,             iptr->dst->prev);
				M_COPY(src->prev->prev,       iptr->dst->prev->prev);
				M_COPY(src->prev->prev->prev, iptr->dst->prev->prev->prev);
				M_COPY(iptr->dst,             iptr->dst->prev->prev->prev->prev);
				M_COPY(iptr->dst->prev,       iptr->dst->prev->prev->prev->prev->prev);
				break;

			case ICMD_SWAP:       /* ..., a, b ==> ..., b, a                      */
				M_COPY(src,       iptr->dst->prev);
				M_COPY(src->prev, iptr->dst);
				break;

		/* integer operations *************************************************/
			case ICMD_INT2BYTE:   /* ..., value  ==> ..., value                   */
				load_var_to_reg_int(s1, src, REG_ITMP1);
				d = reg_of_var(rd, iptr->dst, REG_ITMP3);
				M_MOV(d, REG_LSL(s1, 24));
				M_MOV(d, REG_ASR(d, 24));
				store_reg_to_var_int(iptr->dst, d);
				break;
			case ICMD_INT2CHAR:   /* ..., value  ==> ..., value                   */
				load_var_to_reg_int(s1, src, REG_ITMP1);
				d = reg_of_var(rd, iptr->dst, REG_ITMP3);
				M_MOV(d, REG_LSL(s1, 16));
				M_MOV(d, REG_LSR(d, 16)); /* ATTENTION: char is unsigned */
				store_reg_to_var_int(iptr->dst, d);
				break;
			case ICMD_INT2SHORT:  /* ..., value  ==> ..., value                   */
				load_var_to_reg_int(s1, src, REG_ITMP1);
				d = reg_of_var(rd, iptr->dst, REG_ITMP3);
				M_MOV(d, REG_LSL(s1, 16));
				M_MOV(d, REG_ASR(d, 16));
				store_reg_to_var_int(iptr->dst, d);
				break;

			case ICMD_I2L:        /* ..., value  ==> ..., value                   */
				load_var_to_reg_int(s1, src, REG_ITMP3);
				d = reg_of_var(rd, iptr->dst, PACK_REGS(REG_ITMP1, REG_ITMP2));
				M_INTMOVE(s1, GET_LOW_REG(d));
				M_MOV(GET_HIGH_REG(d), REG_ASR(s1, 32));
				/*M_BREAKPOINT(0x1111);*/
				store_reg_to_var_int(iptr->dst, d);
				break;
			case ICMD_L2I:        /* ..., value  ==> ..., value                   */
				load_var_to_reg_int(s1, src, PACK_REGS(REG_ITMP1, REG_ITMP2));
				d = reg_of_var(rd, iptr->dst, REG_ITMP3);
				/* TODO: correct this! */
				M_INTMOVE(GET_LOW_REG(s1), d);
				store_reg_to_var_int(iptr->dst, d);
				break;

			case ICMD_INEG:       /* ..., value  ==> ..., - value                 */
				load_var_to_reg_int(s1, src, REG_ITMP1); 
				d = reg_of_var(rd, iptr->dst, REG_ITMP3);
				M_NEG(d, s1);
				store_reg_to_var_int(iptr->dst, d);
				break;
			case ICMD_LNEG:       /* ..., value  ==> ..., - value                 */
				load_var_to_reg_int(s1, src, PACK_REGS(REG_ITMP1, REG_ITMP2));
				d = reg_of_var(rd, iptr->dst, PACK_REGS(REG_ITMP3, REG_ITMP2));
				M_RSB_IMMS(GET_LOW_REG(d), GET_LOW_REG(s1), 0);
				M_RSC_IMM(GET_HIGH_REG(d), GET_HIGH_REG(s1), 0);
				/*M_BREAKPOINT(0x1111);*/
				store_reg_to_var_int(iptr->dst, d);
				break;

			case ICMD_IADD:       /* ..., val1, val2  ==> ..., val1 + val2        */
				load_var_to_reg_int(s1, src->prev, REG_ITMP1);
				load_var_to_reg_int(s2, src, REG_ITMP2);
				d = reg_of_var(rd, iptr->dst, REG_ITMP3);
				M_ADD(d, s1, s2);
				store_reg_to_var_int(iptr->dst, d);
				break;
			case ICMD_LADD:       /* ..., val1, val2  ==> ..., val1 + val2        */
				load_var_to_reg_int_low(s1, src->prev, REG_ITMP1);
				load_var_to_reg_int_low(s2, src, REG_ITMP2);
				d = reg_of_var(rd, iptr->dst, PACK_REGS(REG_ITMP2, REG_ITMP3));
				M_ADD_S(GET_LOW_REG(d), s1, s2);
				load_var_to_reg_int_high(s1, src->prev, REG_ITMP1);
				load_var_to_reg_int_high(s2, src, REG_ITMP3);
				M_ADC(GET_HIGH_REG(d), s1, s2);
				/*M_BREAKPOINT(0x1111);*/
				store_reg_to_var_int(iptr->dst, d);
				break;

			case ICMD_IADDCONST:  /* ..., value  ==> ..., value + constant        */
			                      /* val.i = constant                             */
				load_var_to_reg_int(s1, src, REG_ITMP1);
				d = reg_of_var(rd, iptr->dst, REG_ITMP3);
				if (IS_IMM(iptr->val.i)) {
					M_ADD_IMM(d, s1, iptr->val.i);
				} else {
					ICONST(REG_ITMP2, iptr->val.i);
					M_ADD(d, s1, REG_ITMP2);
				}
				store_reg_to_var_int(iptr->dst, d);
				break;
			case ICMD_LADDCONST:  /* ..., value  ==> ..., value + constant        */
			                      /* val.l = constant                             */
				s3 = iptr->val.l & 0xffffffff;
				load_var_to_reg_int_low(s1, src, REG_ITMP1);
				d = reg_of_var(rd, iptr->dst, PACK_REGS(REG_ITMP2, REG_ITMP3));
				if (IS_IMM(s3)) {
					M_ADD_IMMS(GET_LOW_REG(d), s1, s3);
				} else {
					ICONST(REG_ITMP2, s3);
					M_ADD_S(GET_LOW_REG(d), s1, REG_ITMP2);
				}
				s3 = iptr->val.l >> 32;
				load_var_to_reg_int_high(s1, src, REG_ITMP1);
				if (IS_IMM(s3)) {
					M_ADC_IMM(GET_HIGH_REG(d), s1, s3);
				} else {
					ICONST(REG_ITMP2, s3);
					M_ADC(GET_HIGH_REG(d), s1, REG_ITMP2);
				}
				break;

			case ICMD_ISUB:       /* ..., val1, val2  ==> ..., val1 - val2        */
				load_var_to_reg_int(s1, src->prev, REG_ITMP1);
				load_var_to_reg_int(s2, src, REG_ITMP2);
				d = reg_of_var(rd, iptr->dst, REG_ITMP3);
				M_SUB(d, s1, s2);
				store_reg_to_var_int(iptr->dst, d);
				break;
			case ICMD_LSUB:       /* ..., val1, val2  ==> ..., val1 - val2        */
				break;

			case ICMD_ISUBCONST:  /* ..., value  ==> ..., value + constant        */
			                      /* val.i = constant                             */
				load_var_to_reg_int(s1, src, REG_ITMP1);
				d = reg_of_var(rd, iptr->dst, REG_ITMP3);
				if (IS_IMM(iptr->val.i)) {
					M_SUB_IMM(d, s1, iptr->val.i);
				} else {
					ICONST(REG_ITMP2, iptr->val.i);
					M_SUB(d, s1, REG_ITMP2);
				}
				store_reg_to_var_int(iptr->dst, d);
				break;
			case ICMD_LSUBCONST:  /* ..., value  ==> ..., value - constant        */
			                      /* val.l = constant                             */
				s3 = iptr->val.l & 0xffffffff;
				load_var_to_reg_int_low(s1, src, REG_ITMP1);
				d = reg_of_var(rd, iptr->dst, PACK_REGS(REG_ITMP2, REG_ITMP3));
				if (IS_IMM(s3)) {
					M_SUB_IMMS(GET_LOW_REG(d), s1, s3);
				} else {
					ICONST(REG_ITMP2, s3);
					M_SUB_S(GET_LOW_REG(d), s1, REG_ITMP2);
				}
				s3 = iptr->val.l >> 32;
				load_var_to_reg_int_high(s1, src, REG_ITMP1);
				if (IS_IMM(s3)) {
					M_SBC_IMM(GET_HIGH_REG(d), s1, s3);
				} else {
					ICONST(REG_ITMP2, s3);
					M_SBC(GET_HIGH_REG(d), s1, REG_ITMP2);
				}
				break;

			case ICMD_IMUL:       /* ..., val1, val2  ==> ..., val1 * val2        */
				load_var_to_reg_int(s1, src->prev, REG_ITMP1);
				load_var_to_reg_int(s2, src, REG_ITMP2);
				d = reg_of_var(rd, iptr->dst, REG_ITMP3);
				M_MUL(d, s1, s2);
				store_reg_to_var_int(iptr->dst, d);
				break;

			case ICMD_IMULPOW2:   /* ..., value  ==> ..., value * (2 ^ constant)  */
			                      /* val.i = constant                             */
				load_var_to_reg_int(s1, src, REG_ITMP2);
				d = reg_of_var(rd, iptr->dst, REG_ITMP3);
				M_MOV(d, REG_LSL(s1, iptr->val.i));
				store_reg_to_var_int(iptr->dst, d);
				break;
			case ICMD_IDIVPOW2:   /* ..., value  ==> ..., value / (2 ^ constant)  */
			                      /* val.i = constant                             */
				load_var_to_reg_int(s1, src, REG_ITMP2);
				d = reg_of_var(rd, iptr->dst, REG_ITMP3);
				M_MOV(d, REG_ASR(s1, iptr->val.i));
				store_reg_to_var_int(iptr->dst, d);
				break;

			case ICMD_ISHL:       /* ..., val1, val2  ==> ..., val1 << val2       */

				load_var_to_reg_int(s1, src->prev, REG_ITMP1);
				load_var_to_reg_int(s2, src, REG_ITMP2);
				d = reg_of_var(rd, iptr->dst, REG_ITMP3);
				/*M_AND_IMM(REG_ITMP3, s2, 0x1f);*/
				/*M_SLL(d, s1, REG_ITMP3);*/
				M_MOV(d, REG_LSL_REG(s1, s2));
				/*M_BREAKPOINT(0x1111);*/
				store_reg_to_var_int(iptr->dst, d);
				break;

			case ICMD_ISHLCONST:  /* ..., value  ==> ..., value << constant       */
			                      /* val.i = constant                             */
				load_var_to_reg_int(s1, src, REG_ITMP1);
				d = reg_of_var(rd, iptr->dst, REG_ITMP3);
				M_MOV(d, REG_LSL(s1, iptr->val.i & 0x1f));
				store_reg_to_var_int(iptr->dst, d);
				break;

			case ICMD_ISHRCONST:  /* ..., value  ==> ..., value >> constant       */
			                      /* val.i = constant                             */
				load_var_to_reg_int(s1, src, REG_ITMP1);
				d = reg_of_var(rd, iptr->dst, REG_ITMP3);
				M_MOV(d, REG_ASR(s1, iptr->val.i & 0x1f));
				store_reg_to_var_int(iptr->dst, d);
				break;

			case ICMD_IUSHRCONST: /* ..., value  ==> ..., value >>> constant     */
			                      /* val.i = constant                             */
				load_var_to_reg_int(s1, src, REG_ITMP1);
				d = reg_of_var(rd, iptr->dst, REG_ITMP3);
				M_MOV(d, REG_LSR(s1, iptr->val.i & 0x1f));
				store_reg_to_var_int(iptr->dst, d);
				break;

			case ICMD_IAND:       /* ..., val1, val2  ==> ..., val1 & val2        */
				load_var_to_reg_int(s1, src->prev, REG_ITMP1);
				load_var_to_reg_int(s2, src, REG_ITMP2);
				d = reg_of_var(rd, iptr->dst, REG_ITMP3);
				M_AND(d, s1, s2);
				store_reg_to_var_int(iptr->dst, d);
				break;
			case ICMD_LAND:       /* ..., val1, val2  ==> ..., val1 & val2        */
				load_var_to_reg_int_low(s1, src->prev, REG_ITMP1);
				load_var_to_reg_int_low(s2, src, REG_ITMP2);
				d = reg_of_var(rd, iptr->dst, PACK_REGS(REG_ITMP2, REG_ITMP3));
				M_AND(GET_LOW_REG(d), s1, s2);
				load_var_to_reg_int_high(s1, src->prev, REG_ITMP1);
				load_var_to_reg_int_high(s2, src, REG_ITMP3);
				M_AND(GET_HIGH_REG(d), s1, s2);
				store_reg_to_var_int(iptr->dst, d);
				break;

			case ICMD_IOR:        /* ..., val1, val2  ==> ..., val1 | val2        */
				load_var_to_reg_int(s1, src->prev, REG_ITMP1);
				load_var_to_reg_int(s2, src, REG_ITMP2);
				d = reg_of_var(rd, iptr->dst, REG_ITMP3);
				M_ORR(d, s1, s2);
				store_reg_to_var_int(iptr->dst, d);
				break;
			case ICMD_LOR:       /* ..., val1, val2  ==> ..., val1 | val2        */	
				load_var_to_reg_int_low(s1, src->prev, REG_ITMP1);
				load_var_to_reg_int_low(s2, src, REG_ITMP2);
				d = reg_of_var(rd, iptr->dst, PACK_REGS(REG_ITMP2, REG_ITMP3));
				M_ORR(GET_LOW_REG(d), s1, s2);
				load_var_to_reg_int_high(s1, src->prev, REG_ITMP1);
				load_var_to_reg_int_high(s2, src, REG_ITMP3);
				M_ORR(GET_HIGH_REG(d), s1, s2);
				store_reg_to_var_int(iptr->dst, d);
				break;

			case ICMD_IXOR:       /* ..., val1, val2  ==> ..., val1 ^ val2        */
				load_var_to_reg_int(s1, src->prev, REG_ITMP1);
				load_var_to_reg_int(s2, src, REG_ITMP2);
				d = reg_of_var(rd, iptr->dst, REG_ITMP3);
				M_EOR(d, s1, s2);
				store_reg_to_var_int(iptr->dst, d);
				break;
			case ICMD_LXOR:       /* ..., val1, val2  ==> ..., val1 ^ val2        */
				load_var_to_reg_int_low(s1, src->prev, REG_ITMP1);
				load_var_to_reg_int_low(s2, src, REG_ITMP2);
				d = reg_of_var(rd, iptr->dst, PACK_REGS(REG_ITMP2, REG_ITMP3));
				M_EOR(GET_LOW_REG(d), s1, s2);
				load_var_to_reg_int_high(s1, src->prev, REG_ITMP1);
				load_var_to_reg_int_high(s2, src, REG_ITMP3);
				M_EOR(GET_HIGH_REG(d), s1, s2);
				store_reg_to_var_int(iptr->dst, d);
				break;


			case ICMD_IINC:       /* ..., value  ==> ..., value + constant        */
			                      /* op1 = variable, val.i = constant             */
				var = &(rd->locals[iptr->op1][TYPE_INT]);
				if (var->flags & INMEMORY) {
					s1 = REG_ITMP1;
					M_LDR(s1, REG_SP, 4 * var->regoff);
				} else
					s1 = var->regoff;

				if (IS_IMM(iptr->val.i)) {
					M_ADD_IMM(s1, s1, iptr->val.i);
				} else if (IS_IMM(-iptr->val.i)) {
					M_SUB_IMM(s1, s1, (-iptr->val.i));
				} else {
					ICONST(REG_ITMP2, iptr->val.i);
					M_ADD(s1, s1, REG_ITMP2);
				}

				if (var->flags & INMEMORY) {
					M_STR(s1, REG_SP, 4 * var->regoff);
				}
				break;

		/* floating operations ************************************************/
			case ICMD_FADD:       /* ..., val1, val2  ==> ..., val1 + val2        */
				load_var_to_reg_flt(s1, src->prev, REG_FTMP1);
				load_var_to_reg_flt(s2, src, REG_FTMP2);
				d = reg_of_var(rd, iptr->dst, REG_FTMP1);
				M_ADFS(d, s1, s2);
				store_reg_to_var_flt(iptr->dst, d);
				break;
			case ICMD_FSUB:       /* ..., val1, val2  ==> ..., val1 - val2        */
				load_var_to_reg_flt(s1, src->prev, REG_FTMP1);
				load_var_to_reg_flt(s2, src, REG_FTMP2);
				d = reg_of_var(rd, iptr->dst, REG_FTMP1);
				M_SUFS(d, s1, s2);
				store_reg_to_var_flt(iptr->dst, d);
				break;
			case ICMD_FMUL:       /* ..., val1, val2  ==> ..., val1 * val2        */
				load_var_to_reg_flt(s1, src->prev, REG_FTMP1);
				load_var_to_reg_flt(s2, src, REG_FTMP2);
				d = reg_of_var(rd, iptr->dst, REG_FTMP1);
				M_MUFS(d, s1, s2);
				store_reg_to_var_flt(iptr->dst, d);
				break;
			case ICMD_FDIV:       /* ..., val1, val2  ==> ..., val1 / val2        */
				load_var_to_reg_flt(s1, src->prev, REG_FTMP1);
				load_var_to_reg_flt(s2, src, REG_FTMP2);
				d = reg_of_var(rd, iptr->dst, REG_FTMP1);
				M_DVFS(d, s1, s2);
				store_reg_to_var_flt(iptr->dst, d);
				break;
			case ICMD_FREM:       /* ..., val1, val2  ==> ..., val1 % val2        */
				load_var_to_reg_flt(s1, src->prev, REG_FTMP1);
				load_var_to_reg_flt(s2, src, REG_FTMP2);
				d = reg_of_var(rd, iptr->dst, REG_FTMP1);
				M_RMFS(d, s1, s2);
				store_reg_to_var_flt(iptr->dst, d);
				break;

			case ICMD_DADD:       /* ..., val1, val2  ==> ..., val1 + val2        */
				load_var_to_reg_flt(s1, src->prev, REG_FTMP1);
				load_var_to_reg_flt(s2, src, REG_FTMP2);
				d = reg_of_var(rd, iptr->dst, REG_FTMP1);
				M_ADFD(d, s1, s2);
				store_reg_to_var_flt(iptr->dst, d);
				break;
			case ICMD_DSUB:       /* ..., val1, val2  ==> ..., val1 - val2        */
				load_var_to_reg_flt(s1, src->prev, REG_FTMP1);
				load_var_to_reg_flt(s2, src, REG_FTMP2);
				d = reg_of_var(rd, iptr->dst, REG_FTMP1);
				M_SUFD(d, s1, s2);
				store_reg_to_var_flt(iptr->dst, d);
				break;
			case ICMD_DMUL:       /* ..., val1, val2  ==> ..., val1 * val2        */
				load_var_to_reg_flt(s1, src->prev, REG_FTMP1);
				load_var_to_reg_flt(s2, src, REG_FTMP2);
				d = reg_of_var(rd, iptr->dst, REG_FTMP1);
				M_MUFD(d, s1, s2);
				store_reg_to_var_flt(iptr->dst, d);
				break;
			case ICMD_DDIV:       /* ..., val1, val2  ==> ..., val1 / val2        */
				load_var_to_reg_flt(s1, src->prev, REG_FTMP1);
				load_var_to_reg_flt(s2, src, REG_FTMP2);
				d = reg_of_var(rd, iptr->dst, REG_FTMP1);
				M_DVFD(d, s1, s2);
				store_reg_to_var_flt(iptr->dst, d);
				break;
			case ICMD_DREM:       /* ..., val1, val2  ==> ..., val1 % val2        */
				load_var_to_reg_flt(s1, src->prev, REG_FTMP1);
				load_var_to_reg_flt(s2, src, REG_FTMP2);
				d = reg_of_var(rd, iptr->dst, REG_FTMP1);
				M_RMFD(d, s1, s2);
				store_reg_to_var_flt(iptr->dst, d);
				break;

			case ICMD_I2F:       /* ..., value  ==> ..., (float) value            */
				load_var_to_reg_int(s1, src, REG_ITMP1);
				d = reg_of_var(rd, iptr->dst, REG_FTMP1);
				M_FLTS(d,s1);
				store_reg_to_var_flt(iptr->dst, d);
				break;
			case ICMD_I2D:       /* ..., value  ==> ..., (double) value           */
				load_var_to_reg_int(s1, src, REG_ITMP1);
				d = reg_of_var(rd, iptr->dst, REG_FTMP1);
				M_FLTD(d,s1);
				store_reg_to_var_flt(iptr->dst, d);
				break;
			case ICMD_F2I:       /* ..., value  ==> ..., (int) value              */
			case ICMD_D2I:
				load_var_to_reg_flt(s1, src, REG_FTMP1);
				d = reg_of_var(rd, iptr->dst, REG_ITMP1);
				M_FIX(d,s1);
				store_reg_to_var_int(iptr->dst, d);
				break;

			case ICMD_D2F:       /* ..., value  ==> ..., (float) value            */
				load_var_to_reg_flt(s1, src, REG_FTMP1);
				d = reg_of_var(rd, iptr->dst, REG_FTMP2);
				M_MVFS(d,s1);
				store_reg_to_var_flt(iptr->dst, d);
				break;
			case ICMD_F2D:       /* ..., value  ==> ..., (double) value           */
				load_var_to_reg_flt(s1, src, REG_FTMP1);
				d = reg_of_var(rd, iptr->dst, REG_FTMP2);
				M_MVFD(d,s1);
				store_reg_to_var_flt(iptr->dst, d);
				break;

			case ICMD_FCMPL:      /* ..., val1, val2  ==> ..., val1 fcmpl val2    */
			case ICMD_DCMPL:
				load_var_to_reg_flt(s1, src->prev, REG_FTMP1);
				load_var_to_reg_flt(s2, src, REG_FTMP2);
				d = reg_of_var(rd, iptr->dst, REG_ITMP3);
				/*M_BREAKPOINT(0x1111);*/
				M_CMF(s2, s1);
				ICONST(d, 0);
				M_SUBGT_IMM(d, d, 1);
				M_ADDLT_IMM(d, d, 1);
				store_reg_to_var_int(iptr->dst, d);
				break;

			case ICMD_FCMPG:      /* ..., val1, val2  ==> ..., val1 fcmpg val2    */
			case ICMD_DCMPG:
				load_var_to_reg_flt(s1, src->prev, REG_FTMP1);
				load_var_to_reg_flt(s2, src, REG_FTMP2);
				d = reg_of_var(rd, iptr->dst, REG_ITMP3);
				/*M_BREAKPOINT(0x1111);*/
				M_CMF(s1, s2);
				ICONST(d, 0);
				M_SUBLT_IMM(d, d, 1);
				M_ADDGT_IMM(d, d, 1);
				store_reg_to_var_int(iptr->dst, d);
				break;

		/* memory operations **************************************************/
			case ICMD_ARRAYLENGTH: /* ..., arrayref  ==> ..., length              */
				load_var_to_reg_int(s1, src, REG_ITMP1);
				d = reg_of_var(rd, iptr->dst, REG_ITMP2);
				gen_nullptr_check(s1);
				M_LDR(d, s1, OFFSET(java_arrayheader, size));
				store_reg_to_var_int(iptr->dst, d);
				break;

			case ICMD_AALOAD:     /* ..., arrayref, index  ==> ..., value         */
				load_var_to_reg_int(s1, src->prev, REG_ITMP1);
				load_var_to_reg_int(s2, src, REG_ITMP2);
				d = reg_of_var(rd, iptr->dst, REG_ITMP3);
				if (iptr->op1 == 0) {
					gen_nullptr_check(s1);
					gen_bound_check(s1, s2);
				}
				/*M_BREAKPOINT(0x3333);*/
				/*ASM_DEBUG_PREPARE;
				M_MOV(0, s1);
				M_MOV(1, s2);
				ASM_DEBUG_EXECUTE;*/

				M_ADD(REG_ITMP1, s1, REG_LSL(s2, 2)); /* REG_ITMP1 = s1 + 4 * s2 */

				/*ASM_DEBUG_PREPARE;
				M_MOV(0, REG_ITMP1);
				ASM_DEBUG_EXECUTE;*/

				M_LDR(d, REG_ITMP1, OFFSET(java_objectarray, data[0]));
				store_reg_to_var_int(iptr->dst, d);
				break;
			case ICMD_IALOAD:     /* ..., arrayref, index  ==> ..., value         */
				load_var_to_reg_int(s1, src->prev, REG_ITMP1);
				load_var_to_reg_int(s2, src, REG_ITMP2);
				d = reg_of_var(rd, iptr->dst, REG_ITMP3);
				if (iptr->op1 == 0) {
					gen_nullptr_check(s1);
					gen_bound_check(s1, s2);
				}
				M_ADD(REG_ITMP1, s1, REG_LSL(s2, 2)); /* REG_ITMP1 = s1 + 4 * s2 */
				M_LDR(d, REG_ITMP1, OFFSET(java_intarray, data[0]));
				store_reg_to_var_int(iptr->dst, d);
				break;
			case ICMD_BALOAD:     /* ..., arrayref, index  ==> ..., value         */
				load_var_to_reg_int(s1, src->prev, REG_ITMP1);
				load_var_to_reg_int(s2, src, REG_ITMP2);
				d = reg_of_var(rd, iptr->dst, REG_ITMP3);
				if (iptr->op1 == 0) {
					gen_nullptr_check(s1);
					gen_bound_check(s1, s2);
				}
				/*M_BREAKPOINT(0x3333);*/
				M_ADD(REG_ITMP1, s1, s2); /* REG_ITMP1 = s1 + 1 * s2 */
				M_LDRSB(d, REG_ITMP1, OFFSET(java_bytearray, data[0]));
				store_reg_to_var_int(iptr->dst, d);
				break;
			case ICMD_CALOAD:     /* ..., arrayref, index  ==> ..., value         */
				load_var_to_reg_int(s1, src->prev, REG_ITMP1);
				load_var_to_reg_int(s2, src, REG_ITMP2);
				d = reg_of_var(rd, iptr->dst, REG_ITMP3);
				if (iptr->op1 == 0) {
					gen_nullptr_check(s1);
					gen_bound_check(s1, s2);
				}
				M_ADD(REG_ITMP1, s1, REG_LSL(s2, 1)); /* REG_ITMP1 = s1 + 2 * s2 */
				M_LDRH(d, REG_ITMP1, OFFSET(java_chararray, data[0]));
				store_reg_to_var_int(iptr->dst, d);
				break;

			case ICMD_BASTORE:    /* ..., arrayref, index, value  ==> ...         */
				load_var_to_reg_int(s1, src->prev->prev, REG_ITMP1);
				load_var_to_reg_int(s2, src->prev, REG_ITMP2);
				if (iptr->op1 == 0) {
					gen_nullptr_check(s1);
					gen_bound_check(s1, s2);
				}
				load_var_to_reg_int(s3, src, REG_ITMP3);
				M_ADD(REG_ITMP1, s1, s2); /* REG_ITMP1 = s1 + 1 * s2 */
				M_STRB(s3, REG_ITMP1, OFFSET(java_bytearray, data[0]));
				/*M_BREAKPOINT(0x1111);*/
				break;
			case ICMD_CASTORE:    /* ..., arrayref, index, value  ==> ...         */
				load_var_to_reg_int(s1, src->prev->prev, REG_ITMP1);
				load_var_to_reg_int(s2, src->prev, REG_ITMP2);
				if (iptr->op1 == 0) {
					gen_nullptr_check(s1);
					gen_bound_check(s1, s2);
				}
				load_var_to_reg_int(s3, src, REG_ITMP3);
				M_ADD(REG_ITMP1, s1, REG_LSL(s2, 1)); /* REG_ITMP1 = s1 + 2 * s2 */
				M_STRH(s3, REG_ITMP1, OFFSET(java_chararray, data[0]));
				/*M_BREAKPOINT(0x1111);*/
				break;

			case ICMD_GETSTATIC:  /* ...  ==> ..., value                          */
			                      /* op1 = type, val.a = field address            */
				if (!iptr->val.a) {
					codegen_addpatchref(cd, mcodeptr,
					                    PATCHER_get_putstatic,
					                    (unresolved_field *) iptr->target);
					if (opt_showdisassemble) M_NOP;
					a = 0;
				} else {
					fieldinfo *fi = iptr->val.a;
					if (!fi->class->initialized) {
						codegen_addpatchref(cd, mcodeptr,
						                    PATCHER_clinit, fi->class);
						if (opt_showdisassemble) M_NOP;
					}
					a = (ptrint) &(fi->value);
				}

				a = dseg_addaddress(cd, a);
				M_LDR(REG_ITMP1, REG_IP, a);
				switch (iptr->op1) {
				case TYPE_INT:
				case TYPE_ADR:
					d = reg_of_var(rd, iptr->dst, REG_ITMP2);
					M_LDR(d, REG_ITMP1, 0);
					store_reg_to_var_int(iptr->dst, d);
					break;
				case TYPE_LNG:
					d = reg_of_var(rd, iptr->dst, PACK_REGS(REG_ITMP2,REG_ITMP3));
					M_LDR(GET_LOW_REG(d), REG_ITMP1, 0);
					M_LDR(GET_HIGH_REG(d), REG_ITMP1, 4);
					store_reg_to_var_int(iptr->dst, d);
					break;
				case TYPE_FLT:
					d = reg_of_var(rd, iptr->dst, REG_FTMP1);
					M_LDFS(d, REG_ITMP1, 0);
					store_reg_to_var_flt(iptr->dst, d);
					break;
				case TYPE_DBL:
					d = reg_of_var(rd, iptr->dst, REG_FTMP1);
					M_LDFD(d, REG_ITMP1, 0);
					store_reg_to_var_flt(iptr->dst, d);
					break;
				}
				break;

			case ICMD_PUTSTATIC:  /* ..., value  ==> ...                          */
			                      /* op1 = type, val.a = field address            */
				if (!iptr->val.a) {
					codegen_addpatchref(cd, mcodeptr,
					                    PATCHER_get_putstatic,
					                    (unresolved_field*) iptr->target);
					if (opt_showdisassemble) M_NOP;
					a = 0;
				} else {
					fieldinfo *fi = iptr->val.a;
					if (!fi->class->initialized) {
						codegen_addpatchref(cd, mcodeptr,
						                    PATCHER_clinit, fi->class);
						if (opt_showdisassemble) M_NOP;
					}
					a = (ptrint) &(fi->value);
				}

				a = dseg_addaddress(cd, a);
				M_LDR(REG_ITMP1, REG_IP, a);
				switch (iptr->op1) {
				case TYPE_INT:
				case TYPE_ADR:
					load_var_to_reg_int(s2, src, REG_ITMP2);
					M_STR(s2, REG_ITMP1, 0);
					break;
				case TYPE_LNG:
					load_var_to_reg_int(s2, src, PACK_REGS(REG_ITMP2,REG_ITMP3));
					M_STR(GET_LOW_REG(s2), REG_ITMP1, 0);
					M_STR(GET_HIGH_REG(s2), REG_ITMP1, 4);
					break;
				case TYPE_FLT:
					load_var_to_reg_flt(s2, src, REG_FTMP1);
					M_STFS(s2, REG_ITMP1, 0);
					break;
				case TYPE_DBL:
					load_var_to_reg_flt(s2, src, REG_FTMP1);
					M_STFD(s2, REG_ITMP1, 0);
					break;
				}
				break;

			case ICMD_GETFIELD:   /* ..., objectref, value  ==> ...               */
			                      /* op1 = type, val.i = field offset             */
				load_var_to_reg_int(s1, src, REG_ITMP1);
				gen_nullptr_check(s1);
				if (!iptr->val.a) {
					codegen_addpatchref(cd, mcodeptr,
					                    PATCHER_get_putfield,
					                    (unresolved_field*) iptr->target);
					if (opt_showdisassemble) M_NOP;
					a = 0;
				} else {
					a = ((fieldinfo*) (iptr->val.a))->offset;
				}
				switch (iptr->op1) {
				case TYPE_INT:
				case TYPE_ADR:
					d = reg_of_var(rd, iptr->dst, REG_ITMP2);
					M_LDR(d, s1, a);
					store_reg_to_var_int(iptr->dst, d);
					break;
				case TYPE_LNG:
					d = reg_of_var(rd, iptr->dst, PACK_REGS(REG_ITMP2,REG_ITMP3));
					M_LDR(GET_LOW_REG(d), s1, a);
					M_LDR(GET_HIGH_REG(d), s1, a+4);
					break;
				case TYPE_FLT:
					d = reg_of_var(rd, iptr->dst, REG_FTMP1);
					M_LDFS(d, s1, a);
					store_reg_to_var_flt(iptr->dst, d);
					break;
				case TYPE_DBL:
					d = reg_of_var(rd, iptr->dst, REG_FTMP1);
					M_LDFD(d, s1, a);
					store_reg_to_var_flt(iptr->dst, d);
					break;
				}
				break;

			case ICMD_PUTFIELD:   /* ..., objectref, value  ==> ...               */
			                      /* op1 = type, val.a = field address            */
				load_var_to_reg_int(s1, src->prev, REG_ITMP1);
				gen_nullptr_check(s1);
				if (!iptr->val.a) {
						codegen_addpatchref(cd, mcodeptr,
											PATCHER_get_putfield,
											(unresolved_field*) iptr->target);
						if (opt_showdisassemble) M_NOP;
						a = 0;
				} else {
						a = ((fieldinfo*) (iptr->val.a))->offset;
				}
				switch (iptr->op1) {
				case TYPE_INT:
				case TYPE_ADR:
					load_var_to_reg_int(s2, src, REG_ITMP2);
					M_STR(s2, s1, a);
					break;
				case TYPE_LNG:
					load_var_to_reg_int(s2, src, PACK_REGS(REG_ITMP2,REG_ITMP3));
					M_STR(GET_LOW_REG(s2), s1, a);
					M_STR(GET_HIGH_REG(s2), s1, a+4);
					break;
				case TYPE_FLT:
					load_var_to_reg_flt(s2, src, REG_FTMP1);
					M_STFS(s2, s1, a);
					break;
				case TYPE_DBL:
					load_var_to_reg_flt(s2, src, REG_FTMP1);
					M_STFD(s2, s1, a);
					break;
				}
				break;

		/* branch operations **************************************************/
			case ICMD_ATHROW:       /* ..., objectref ==> ... (, objectref)       */
				load_var_to_reg_int(s1, src, REG_ITMP1);
				ICONST(3, 0xffabcd); /* TODO: this is only to make debugging easier */
				a = dseg_addaddress(cd, asm_handle_exception);
				M_INTMOVE(s1, REG_ITMP1_XPTR);
				M_ADD_IMM(REG_ITMP2_XPC, REG_PC, 4);
				M_MOV(REG_LR, REG_PC);
				M_LDR(REG_PC, REG_IP, a);
				M_NOP;              /* nop ensures that XPC is less than the end */
				                    /* of basic block                            */
				break;

			case ICMD_GOTO:         /* ... ==> ...                                */
			                        /* op1 = target JavaVM pc                     */
				M_B(0);
				codegen_addreference(cd, (basicblock*) iptr->target, mcodeptr);
				break;

			case ICMD_JSR:          /* ... ==> ...                                */
			                        /* op1 = target JavaVM pc                     */
				M_MOV(REG_ITMP1, REG_PC);
				M_B(0);
				codegen_addreference(cd, (basicblock *) iptr->target, mcodeptr);
				break;
			
			case ICMD_RET:          /* ... ==> ...                                */
			                        /* op1 = local variable                       */
				var = &(rd->locals[iptr->op1][TYPE_ADR]);
				if (var->flags & INMEMORY) {
					M_LDR(REG_PC, REG_SP, 4 * var->regoff);
				} else {
					M_MOV(REG_PC, var->regoff);
				}
				break;

			case ICMD_IFNULL:       /* ..., value ==> ...                         */
			                        /* op1 = target JavaVM pc                     */
				load_var_to_reg_int(s1, src, REG_ITMP1);
				M_TEQ_IMM(s1, 0);
				M_BEQ(0);
				codegen_addreference(cd, (basicblock*) iptr->target, mcodeptr);
				break;

			case ICMD_IFNONNULL:    /* ..., value ==> ...                         */
			                        /* op1 = target JavaVM pc                     */
				load_var_to_reg_int(s1, src, REG_ITMP1);
				M_TEQ_IMM(s1, 0);
				M_BNE(0);
				codegen_addreference(cd, (basicblock*) iptr->target, mcodeptr);
				break;

			case ICMD_IFLT:         /* ..., value ==> ...                         */
			case ICMD_IFLE:         /* op1 = target JavaVM pc, val.i = constant   */
			case ICMD_IFGT:
			case ICMD_IFGE:
			case ICMD_IFEQ:
			case ICMD_IFNE:
				load_var_to_reg_int(s1, src, REG_ITMP1);
				M_COMPARE(s1, iptr->val.i, UNCOND);
				switch(iptr->opc) {
					case ICMD_IFLT: M_BLT(0); break;
					case ICMD_IFLE: M_BLE(0); break;
					case ICMD_IFGT: M_BGT(0); break;
					case ICMD_IFGE: M_BGE(0); break;
					case ICMD_IFEQ: M_BEQ(0); break;
					case ICMD_IFNE: M_BNE(0); break;
					default: /*assert*/;
				}
				codegen_addreference(cd, (basicblock*) iptr->target, mcodeptr);
				break;

			case ICMD_IF_LLT:       /* ..., value ==> ...                         */
			case ICMD_IF_LLE:       /* op1 = target JavaVM pc, val.l = constant   */
			case ICMD_IF_LGT:
			case ICMD_IF_LGE:
			case ICMD_IF_LEQ:
			case ICMD_IF_LNE:
				load_var_to_reg_int_low(s1, src, REG_ITMP1);
				load_var_to_reg_int_high(s2, src, REG_ITMP3);
				M_COMPARE(s2, (iptr->val.l >> 32), UNCOND);
				switch(iptr->opc) {
					case ICMD_IF_LLT: M_COMPARE(s1, (iptr->val.l & 0xffffffff), COND_EQ); M_BLT(0); break;
					case ICMD_IF_LLE: M_COMPARE(s1, (iptr->val.l & 0xffffffff), COND_EQ); M_BLE(0); break;
					case ICMD_IF_LGT: M_COMPARE(s1, (iptr->val.l & 0xffffffff), COND_EQ); M_BGT(0); break;
					case ICMD_IF_LGE: M_COMPARE(s1, (iptr->val.l & 0xffffffff), COND_EQ); M_BGE(0); break;
					case ICMD_IF_LEQ: M_COMPARE(s1, (iptr->val.l & 0xffffffff), COND_EQ); M_BEQ(0); break;
					case ICMD_IF_LNE: M_COMPARE(s1, (iptr->val.l & 0xffffffff), COND_EQ); M_BNE(0); break;
					default: /*assert*/;
				}
				codegen_addreference(cd, (basicblock*) iptr->target, mcodeptr);
				break;

			case ICMD_IF_ICMPEQ:    /* ..., value, value ==> ...                  */
			case ICMD_IF_ACMPEQ:    /* op1 = target JavaVM pc                     */
			case ICMD_IF_ICMPNE:
			case ICMD_IF_ACMPNE:
			case ICMD_IF_ICMPLT:
			case ICMD_IF_ICMPLE:
			case ICMD_IF_ICMPGT:
			case ICMD_IF_ICMPGE:
				load_var_to_reg_int(s1, src->prev, REG_ITMP1);
				load_var_to_reg_int(s2, src, REG_ITMP2);
				M_CMP(s1, s2);
				switch(iptr->opc) {
					case ICMD_IF_ICMPLT: M_BLT(0); break;
					case ICMD_IF_ICMPLE: M_BLE(0); break;
					case ICMD_IF_ICMPGT: M_BGT(0); break;
					case ICMD_IF_ICMPGE: M_BGE(0); break;
					case ICMD_IF_ICMPEQ: /* fall through */
					case ICMD_IF_ACMPEQ: M_BEQ(0); break;
					case ICMD_IF_ICMPNE: /* fall through */
					case ICMD_IF_ACMPNE: M_BNE(0); break;
					default: /*assert*/;
				}
				codegen_addreference(cd, (basicblock *) iptr->target, mcodeptr);
				break;

			case ICMD_IF_LCMPEQ:    /* ..., value, value ==> ...                  */
			case ICMD_IF_LCMPNE:    /* op1 = target JavaVM pc                     */
			case ICMD_IF_LCMPLT:
			case ICMD_IF_LCMPLE:
			case ICMD_IF_LCMPGT:
			case ICMD_IF_LCMPGE:
				load_var_to_reg_int_high(s1, src->prev, REG_ITMP1);
				load_var_to_reg_int_high(s2, src, REG_ITMP2);
				M_CMP(s1, s2);
				load_var_to_reg_int_low(s1, src->prev, REG_ITMP1);
				load_var_to_reg_int_low(s2, src, REG_ITMP2);
				switch(iptr->opc) {
					case ICMD_IF_LCMPEQ: M_DAT(COND_EQ,0x0a,0,s1,1,0,s2); M_BEQ(0); break;
					case ICMD_IF_LCMPNE: M_DAT(COND_EQ,0x0a,0,s1,1,0,s2); M_BNE(0); break;
					case ICMD_IF_LCMPLT: M_DAT(COND_EQ,0x0a,0,s1,1,0,s2); M_BLT(0); break;
					case ICMD_IF_LCMPLE: M_DAT(COND_EQ,0x0a,0,s1,1,0,s2); M_BLE(0); break;
					case ICMD_IF_LCMPGT: M_DAT(COND_EQ,0x0a,0,s1,1,0,s2); M_BGT(0); break;
					case ICMD_IF_LCMPGE: M_DAT(COND_EQ,0x0a,0,s1,1,0,s2); M_BGE(0); break;
					default: /*assert*/;
				}
				codegen_addreference(cd, (basicblock *) iptr->target, mcodeptr);
				break;

			case ICMD_TABLESWITCH:  /* ..., index ==> ...                         */
				{
				s4 j, *s4ptr;
				void **tptr;

				tptr = (void **) iptr->target;
				s4ptr = iptr->val.a;
				j = s4ptr[1]; /* low  */
				i = s4ptr[2]; /* high */

				/* calculate new index (index - low) */
				load_var_to_reg_int(s1, src, REG_ITMP1);
				if (j  == 0) {
					M_INTMOVE(s1, REG_ITMP1);
				} else if (IS_IMM(j)) {
					M_SUB_IMM(REG_ITMP1, s1, j);
				} else {
					ICONST(REG_ITMP2, j);
					M_SUB(REG_ITMP1, s1, REG_ITMP2);
				}

				/* range check (index <= high-low) */
				i = i - j + 1;
				M_COMPARE(REG_ITMP1, i-1, UNCOND);
				M_BHI(0); /* unsigned greater than */
				codegen_addreference(cd, (basicblock*) tptr[0], mcodeptr);

				/* build jump table top down and use address of lowest entry */
				tptr += i;
				while (--i >= 0) {
					dseg_addtarget(cd, (basicblock*) tptr[0]); 
					--tptr;
				}
				}

				/* length of dataseg after last dseg_addtarget is used by load */
				M_ADD(REG_ITMP2, REG_IP, REG_LSL(REG_ITMP1, 2));
				M_LDR(REG_PC, REG_ITMP2, -(cd->dseglen));
				break;

			case ICMD_LOOKUPSWITCH: /* ..., key ==> ...                           */
				{
				s4 j, val, *s4ptr;
				void **tptr;

				tptr = (void **) iptr->target;
				s4ptr = iptr->val.a;
				j = s4ptr[0]; /* default  */
				i = s4ptr[1]; /* count    */

				/* TODO: MCODECHECK!!! */

				/* compare keys */
				load_var_to_reg_int(s1, src, REG_ITMP1);
				while (--i >= 0) {
					s4ptr += 2;
					++tptr;
					val = s4ptr[0];
					M_COMPARE(s1, val, UNCOND);
					M_BEQ(0);
					codegen_addreference(cd, (basicblock*) tptr[0], mcodeptr);
				}

				/* default branch */
				M_B(0);
				tptr = (void **) iptr->target;
				codegen_addreference(cd, (basicblock *) tptr[0], mcodeptr);
	
				}
				break;

			case ICMD_IRETURN:      /* ..., retvalue ==> ...                      */
			case ICMD_ARETURN:
				load_var_to_reg_int(s1, src, REG_RESULT);
				M_INTMOVE(s1, REG_RESULT);
				goto ICMD_RETURN_do;

			case ICMD_LRETURN:      /* ..., retvalue ==> ...                      */
				dolog("!!! RETURN: long not tested yet (%s.%s)", m->class->name->text, m->name->text);
				load_var_to_reg_int(s1, src, PACK_REGS(REG_RESULT, REG_RESULT2));
				M_INTMOVE_TYPED(src->type, s1, PACK_REGS(REG_RESULT, REG_RESULT2));
				goto ICMD_RETURN_do;
			case ICMD_FRETURN:
				load_var_to_reg_flt(s1, src, REG_FTMP1);
				M_FLTSTORE_TYPED(src->type, s1, REG_RESULT);
				goto ICMD_RETURN_do;
			case ICMD_DRETURN:
				dolog("!!! RETURN: double not tested yet (%s.%s)", m->class->name->text, m->name->text);
				load_var_to_reg_flt(s1, src, REG_FTMP1);
				M_FLTSTORE_TYPED(src->type, s1, PACK_REGS(REG_RESULT, REG_RESULT2));
				goto ICMD_RETURN_do;

			/* TODO: do this! */
			case ICMD_RETURN:       /* ...  ==> ...                               */
			ICMD_RETURN_do:
				/* call trace function */
				/* TODO: why does it not work without??? */
				if (runverbose) {
					M_STMFD(0x0f | (1<<REG_LR) | (1<<REG_IP), REG_SP);

					a = dseg_addaddress(cd, m);
					M_MOV(2, 1);
					M_MOV(1, 0);
					M_LDR(0, REG_IP, a);

					M_LONGBRANCH(builtin_displaymethodstop);

					M_LDMFD(0x0f | (1<<REG_LR) | (1<<REG_IP), REG_SP);
				}

#if defined(USE_THREADS)
				log_text("!!! ICMD_RETURN: threads not implemented yet");
#endif

				/* deallocate stackframe for spilled variables */
				if (spilledregs_num) {
					M_ADD_IMM(REG_SP, REG_SP, spilledregs_num * 4);
				}

				/* restore callee saved registers + do return */
				if (savedregs_bitmask) {
					if (!m->isleafmethod) {
						savedregs_bitmask &= ~(1<<REG_LR);
						savedregs_bitmask |= (1<<REG_PC);
					}
					M_LDMFD(savedregs_bitmask, REG_SP);
				}
				/* if LR was not on stack, we need to return manually */
				if (m->isleafmethod) {
					M_MOV(REG_PC,REG_LR);
				}
				break;

			case ICMD_BUILTIN:      /* ..., arg1, arg2, arg3 ==> ...              */
			                        /* op1 = arg count val.a = builtintable entry */
				bte = iptr->val.a;
				md = bte->md;
				goto ICMD_INVOKE_do;

			case ICMD_INVOKESTATIC: /* ..., [arg1, [arg2 ...]] ==> ...            */
			                        /* op1 = arg count, val.a = method pointer    */
			case ICMD_INVOKESPECIAL:/* ..., objectref, [arg1, [arg2 ...]] ==> ... */
			case ICMD_INVOKEVIRTUAL:/* op1 = arg count, val.a = method pointer    */
			case ICMD_INVOKEINTERFACE:
				lm = iptr->val.a;
				if (lm)
					md = lm->parseddesc;
				else {
					unresolved_method *um = iptr->target;
					md = um->methodref->parseddesc.md;
				}

			ICMD_INVOKE_do:
				/* copy arguments to registers or stack location                  */
				s3 = iptr->op1;
				for (s3 = s3 - 1; s3 >= 0; s3--, src = src->prev) {
					/* TODO: what's this? */
					/* TODO: ugly! */
					if (src->varkind == ARGVAR && IS_INT_LNG_TYPE(src->type))
						continue;
					if (IS_INT_LNG_TYPE(src->type)) {
						if (!md->params[s3].inmemory) {
							s1 = ARGUMENT_REGS(src->type, md->params[s3].regoff);
							load_var_to_reg_int(d, src, s1);
							M_INTMOVE_TYPED(src->type, d, s1);
						} else {
							if (IS_2_WORD_TYPE(src->type)) {
								load_var_to_reg_int(d, src, PACK_REGS(REG_ITMP1, REG_ITMP2));
								M_STR(GET_LOW_REG(d), REG_SP, md->params[s3].regoff * 4);
								M_STR(GET_HIGH_REG(d), REG_SP, md->params[s3].regoff * 4 + 4);
							} else {
								load_var_to_reg_int(d, src, REG_ITMP1);
								M_STR(d, REG_SP, md->params[s3].regoff * 4);
							}
							/* TODO: wo kommt der stack-space her? */
							log_text("!!! INVOKE: integer stack arguments not tested yet");
						}
					} else {
						log_text("!!! INVOKE: float and double not tested yet");
						if (!md->params[s3].inmemory) {
							s1 = ARGUMENT_REGS(src->type, md->params[s3].regoff);
							load_var_to_reg_flt(d, src, REG_FTMP1);
							M_FLTSTORE_TYPED(src->type, d, s1);
						} else {
							load_var_to_reg_flt(d, src, REG_FTMP1);
							if (IS_2_WORD_TYPE(src->type)) {
								M_STFD(d, REG_SP, md->params[s3].regoff * 4);
							} else {
								M_STFS(d, REG_SP, md->params[s3].regoff * 4);
							}
							/* TODO: wo kommt der stack-space her? */
							log_text("!!! INVOKE: float stack arguments not tested yet");
						}
					}
				}

				/* invocation specific (target method must be in REG_IP afterwards) */
				switch (iptr->opc) {
				case ICMD_BUILTIN:
					if (iptr->target) {
						codegen_addpatchref(cd, mcodeptr, bte->fp, iptr->target);
						if (opt_showdisassemble) M_NOP;
						a = 0;
					} else {
						a = (ptrint) bte->fp;
					}
					a = dseg_addaddress(cd, a);
					M_LDR(REG_IP, REG_IP, a); /* Pointer to built-in-function */
					break;
				case ICMD_INVOKESPECIAL:
					gen_nullptr_check(rd->argintregs[0]);
					M_LDR(REG_ITMP1, rd->argintregs[0], 0); /* hardware nullptr */
					/* fall through */
				case ICMD_INVOKESTATIC:
					if (!lm) {
						unresolved_method *um = iptr->target;
						codegen_addpatchref(cd, mcodeptr,
						                    PATCHER_invokestatic_special, um);
						if (opt_showdisassemble) M_NOP;
						a = 0;
					} else {
						a = (ptrint) lm->stubroutine;
					}
					a = dseg_addaddress(cd, a);
					M_LDR(REG_IP, REG_IP, a);            /* Pointer to method */
					break;
				case ICMD_INVOKEVIRTUAL:
					gen_nullptr_check(rd->argintregs[0]);
					if (!lm) {
						unresolved_method *um = iptr->target;
						codegen_addpatchref(cd, mcodeptr,
						                    PATCHER_invokevirtual, um);
						if (opt_showdisassemble) M_NOP;
						s1 = 0;
					} else {
						s1 = OFFSET(vftbl_t, table[0]) +
							sizeof(methodptr) * lm->vftblindex;
					}
					M_LDR(REG_METHODPTR, rd->argintregs[0],
					             OFFSET(java_objectheader, vftbl));
					M_LDR(REG_IP, REG_METHODPTR, s1);
					break;
				case ICMD_INVOKEINTERFACE:
					gen_nullptr_check(rd->argintregs[0]);
					if (!lm) {
						unresolved_method *um = iptr->target;
						codegen_addpatchref(cd, mcodeptr,
						                    PATCHER_invokeinterface, um);
						if (opt_showdisassemble) M_NOP;
						s1 = 0;
						s2 = 0;
					} else {
						s1 = OFFSET(vftbl_t, interfacetable[0]) -
							sizeof(methodptr*) * lm->class->index;
						s2 = sizeof(methodptr) * (lm - lm->class->methods);
					}
					M_LDR(REG_METHODPTR, rd->argintregs[0],
					             OFFSET(java_objectheader, vftbl));
					M_LDR(REG_METHODPTR, REG_METHODPTR, s1);
					M_LDR(REG_IP, REG_METHODPTR, s2);
					break;
				}

				/* call debugger */
				if (0 && iptr->opc == ICMD_BUILTIN) {
                                	ASM_DEBUG_PREPARE;
					M_MOV(1, 0);
                                	ICONST(0, 0xaa);
                                	M_MOV(2, REG_IP);
                                	M_MOV(3, REG_SP);
                                	ASM_DEBUG_EXECUTE;
				}

				/* do invokation here */
                                /*M_BREAKPOINT(0xfabc);*/
				M_MOV(REG_LR, REG_PC);           /* save return address in LR */
				M_MOV(REG_PC, REG_IP);           /* branch to method          */
				/*M_BREAKPOINT(0xfabc);*/            /* test if LR was correct    */

				/* recompute ip (pv) */
				/* TODO: this is important!!! */
				s1 = (s4) ((u1*) mcodeptr - cd->mcodebase);
				s1 += 8;                         /* we use PC relative addr.  */
				M_RECOMPUTE_IP(s1);

				/* call debugger */
				if (0 && iptr->opc == ICMD_BUILTIN) {
					ASM_DEBUG_PREPARE;
					M_MOV(1, 0);
					ICONST(0, 0xbb);
					M_MOV(2, REG_IP);
					M_MOV(3, REG_SP);
					ASM_DEBUG_EXECUTE;
				}

				/* d contains return type */
				d = md->returntype.type;
				if (d != TYPE_VOID) {
					if (IS_INT_LNG_TYPE(iptr->dst->type)) {
						/* TODO: improve this */
						if (IS_2_WORD_TYPE(iptr->dst->type)) {
							s1 = reg_of_var(rd, iptr->dst,
							   PACK_REGS(REG_RESULT, REG_RESULT2));
							M_INTMOVE_TYPED(iptr->dst->type,
							   PACK_REGS(REG_RESULT, REG_RESULT2), s1);
						} else {
							s1 = reg_of_var(rd, iptr->dst, REG_RESULT);
							M_INTMOVE(REG_RESULT, s1);
						}
						store_reg_to_var_int(iptr->dst, s1);
					} else {
						s1 = reg_of_var(rd, iptr->dst, REG_FTMP1);
						/* TODO: improve this! (this is really ugly) */
						if (IS_2_WORD_TYPE(iptr->dst->type)) {
							M_FLTLOAD_TYPED(iptr->dst->type,
							   PACK_REGS(REG_RESULT, REG_RESULT2), s1);
						} else {
							M_FLTLOAD_TYPED(iptr->dst->type, REG_RESULT, s1);
						}
						store_reg_to_var_flt(iptr->dst, s1);
					}
				}
				break;

			case ICMD_CHECKCAST:  /* ..., objectref ==> ..., objectref            */
			                      /* op1:   0 == array, 1 == class                */
			                      /* val.a: (classinfo*) superclass               */
				{
				classinfo *super;
				vftbl_t   *supervftbl;
				s4         superindex;

				super = (classinfo*) iptr->val.a;
				if (!super) {
					superindex = 0;
					supervftbl = NULL;
				} else {
					superindex = super->index;
					supervftbl = super->vftbl;
				}
			
#if defined(USE_THREADS) && defined(NATIVE_THREADS)
				log_text("!!! CHECKCAST: threads not implemented yet");
#endif
				load_var_to_reg_int(s1, src, REG_ITMP1);
				log_text("!!! CHECKCAST: not implemented at all");

				/* TODO: implement this! */
				}

				d = reg_of_var(rd, iptr->dst, REG_ITMP3);
				M_INTMOVE(s1, d);
				store_reg_to_var_int(iptr->dst, d);
				break;

			case ICMD_INSTANCEOF: /* ..., objectref ==> ..., intresult            */
			                      /* op1:   0 == array, 1 == class                */
			                      /* val.a: (classinfo*) superclass               */
				load_var_to_reg_int(s1, src, REG_ITMP1);
				log_text("!!! INSTANCEOF: not implemented at all");
				d = reg_of_var(rd, iptr->dst, REG_ITMP2);
				ICONST(d, 1);
				store_reg_to_var_int(iptr->dst, d);
				break;

			case ICMD_MULTIANEWARRAY:/* ..., cnt1, [cnt2, ...] ==> ..., arrayref  */
			                      /* op1 = dimension, val.a = array descriptor    */

				/* check for negative sizes and copy sizes to stack if necessary  */
				/* TODO: MCODECHECK((iptr->op1 << 1) + 64); */
				for (s1 = iptr->op1; --s1 >= 0; src = src->prev) {
					/* check for negative sizes */
					load_var_to_reg_int(s2, src, REG_ITMP1);
					M_CMP_IMM(s2, 0);
					M_BLT(-2);
					codegen_addxcheckarefs(cd, mcodeptr);
					/* copy SAVEDVAR sizes to stack */
					if (src->varkind != ARGVAR) {
						M_STR(s2, REG_SP, s1 * 4);
					}
				}

				/* is patcher function set? */
				if (iptr->target) {
					codegen_addpatchref(cd, mcodeptr,
					                    (functionptr) iptr->target, iptr->val.a);
					if (opt_showdisassemble) M_NOP;
					a = 0;
				} else {
					a = (ptrint) iptr->val.a;
				}

				/* a0 = dimension count */
				ICONST(rd->argintregs[0], iptr->op1);

				/* a1 = arraydescriptor */
				a = dseg_addaddress(cd, a);
				M_LDR(rd->argintregs[1], REG_IP, a);

				/* a2 = pointer to dimensions = stack pointer */
				M_INTMOVE(REG_SP, rd->argintregs[2]);

				/* call builtin_multianewarray here */
				a = dseg_addaddress(cd, (void*) BUILTIN_multianewarray);
				M_MOV(REG_LR, REG_PC);           /* save return address in LR */
				M_LDR(REG_PC, REG_IP, a);        /* branch to method          */

				/* recompute ip (pv) */
				/* TODO: this is important!!! */
				s1 = (s4) ((u1*) mcodeptr - cd->mcodebase);
				s1 += 8;                         /* we use PC relative addr.  */
				M_RECOMPUTE_IP(s1);

				/* get arrayref */
				s1 = reg_of_var(rd, iptr->dst, REG_RESULT);
				M_INTMOVE(REG_RESULT, s1);
				store_reg_to_var_int(iptr->dst, s1);
				break;

			case ICMD_CHECKEXCEPTION:    /* ... ==> ...                           */
				M_TST(REG_RESULT, REG_RESULT);
				M_BEQ(-2);
				codegen_addxexceptionrefs(cd, mcodeptr);
				break;

			case ICMD_CHECKASIZE:  /* ..., size ==> ..., size                     */
				load_var_to_reg_int(s1, src, REG_ITMP1);
				M_CMP_IMM(s1, 0);
				M_BLT(-2);
				codegen_addxcheckarefs(cd, mcodeptr);
				break;

			default:
				/* TODO: clean this up! */
				M_NOP; M_BREAKPOINT(iptr->opc); M_NOP;
				printf("=== Unknown ICMD %s in %s.%s ===\n", icmd_names[iptr->opc], m->class->name->text, m->name->text);
/*				throw_cacao_exception_exit(string_java_lang_InternalError,
				                           "Unknown ICMD %d", iptr->opc);
*/			} /* the big switch */

		} /* for all instructions */

		/* copy values to interface registers */
		/* TODO: implement this! */
		src = bptr->outstack;
		len = bptr->outdepth;
		MCODECHECK(64 + len); /* TODO!!! */
		while (src) {
			/*log_text("!!! OUTSTACK: we have some regs here");*/
			len--;
			if ((src->varkind != STACKVAR)) {
				t = src->type;
				if (IS_FLT_DBL_TYPE(t)) {
					load_var_to_reg_flt(s1, src, REG_FTMP1);
					if (!(rd->interfaces[len][t].flags & INMEMORY)) {
						M_FLTMOVE_TYPED(t, s1, rd->interfaces[len][t].regoff);	
					} else {
						log_text("!!! OUTSTACK: float and double interface problem!!!");
						/*M_DST(s1, REG_SP, rd->interfaces[len][s2].regoff * 4);*/
					}	
				} else {
					load_var_to_reg_int(s1, src, REG_ITMP1);
					if (!(rd->interfaces[len][t].flags & INMEMORY)) {
						M_INTMOVE_TYPED(t, s1, rd->interfaces[len][t].regoff);
					} else {
						if (IS_2_WORD_TYPE(t)) {
							M_STR(GET_HIGH_REG(s1), REG_SP, rd->interfaces[len][t].regoff * 4);
							M_STR(GET_LOW_REG(s1), REG_SP, rd->interfaces[len][t].regoff * 4 + 4);
						} else {
							M_STR(s1, REG_SP, rd->interfaces[len][t].regoff * 4);
						}
					}
				}
			}
			src = src->prev;
		}

	} /* for all basic blocks */

	codegen_createlinenumbertable(cd);

	/* generate bound check stubs */
	/* generate negative array size check stubs */
	/* generate cast check stubs */
	/* generate exception check stubs */
	/* generate null pointer check stubs */
	/* generate patcher stub call code */
	{
	patchref *pref;
	u4        mcode;
	s4       *tmpmcodeptr;
	s4       *xcodeptr;

	for (pref = cd->patchrefs; pref != NULL; pref = pref->next) {
		/* check code segment size */
		MCODECHECK(50); /* TODO */

		/* Get machine code which is patched back in later. The call is   */
		/* 1 instruction word long.                                       */
		xcodeptr = (s4 *) (cd->mcodebase + pref->branchpos);
		mcode = *xcodeptr;

		/* patch in the call to call the following code (done at compile  */
		/* time)                                                          */
		tmpmcodeptr = mcodeptr;         /* save current mcodeptr          */
		mcodeptr = xcodeptr;            /* set mcodeptr to patch position */
		M_B(tmpmcodeptr - (xcodeptr + 1) - 1);
		mcodeptr = tmpmcodeptr;         /* restore the current mcodeptr   */

		/* calculate return address */
		/* ATTENTION: we can not use BL to branch to patcher stub,        */
		/* ATTENTION: because we need to preserve LR for leaf methods     */
		ICONST(REG_ITMP3, (s4) (tmpmcodeptr - (xcodeptr + 1) + 3) * 4);
		M_SUB(REG_ITMP3, REG_PC, REG_ITMP3);

		/* create stack frame */
		M_SUB_IMM(REG_SP, REG_SP, 5*4);

		/* move return address onto stack */
		M_STR(REG_ITMP3, REG_SP, 4*4);

		/* move pointer to java_objectheader onto stack */
#if defined(USE_THREADS) && defined(NATIVE_THREADS)
		log_text("!!! PATCHER: threads not implemented yet");
#else
		M_EOR(REG_ITMP3, REG_ITMP3, REG_ITMP3);
		M_STR(REG_ITMP3, REG_SP, 3*4);
#endif

		/* move machine code onto stack */
		a = dseg_adds4(cd, mcode);
		M_LDR(REG_ITMP3, REG_IP, a);
		M_STR(REG_ITMP3, REG_SP, 2*4);

		/* move class/method/field reference onto stack */
		a = dseg_addaddress(cd, pref->ref);
		M_LDR(REG_ITMP3, REG_IP, a);
		M_STR(REG_ITMP3, REG_SP, 1*4);

		/* move patcher function pointer onto stack */
		a = dseg_addaddress(cd, pref->patcher);
		M_LDR(REG_ITMP3, REG_IP, a);
		M_STR(REG_ITMP3, REG_SP, 0*4);

		/* finally call the patcher via asm_wrapper_patcher */
		/* ATTENTION: don't use REG_IP here, because some patchers need it */
		a = dseg_addaddress(cd, asm_wrapper_patcher);
		M_LDR(REG_PC, REG_IP, a);
	}
	}

	codegen_finish(m, cd, (s4) ((u1*) mcodeptr - cd->mcodebase));

	/* synchronize instruction cache */
	asm_sync_instruction_cache();
}


/* function createcompilerstub *************************************************

   creates a stub routine which calls the compiler

*******************************************************************************/

#define COMPSTUBSIZE 6 /* TODO: correct this value */

functionptr createcompilerstub(methodinfo *m)
{
	s4 *s = CNEW(s4, COMPSTUBSIZE);
	s4 *mcodeptr = s;                   /* code generation pointer            */

	M_SUB_IMM(REG_SP, REG_SP, 4);       /* use stack to pass method pointer   */
	M_LDR(REG_ITMP1, REG_PC, 4);
	M_STR(REG_ITMP1, REG_SP, 0);

	/*M_LDR(REG_ITMP1, REG_PC, 4);*/        /* use ITMP1 to pass method pointer   */
	M_LDR(REG_PC, REG_PC, 0);           /* jump to asm_call_jit_compiler      */
	DCD((s4) m);
	DCD((s4) asm_call_jit_compiler);

#if defined(STATISTICS)
	if (opt_stat)
		count_cstub_len += COMPSTUBSIZE * 4;
#endif

	/* synchronize instruction cache */
	asm_sync_instruction_cache();

	return (functionptr)(ptrint) s;
}


/* createnativestub ************************************************************

   Creates a stub routine which calls a native method.

*******************************************************************************/

functionptr createnativestub(functionptr f, methodinfo *m, codegendata *cd, registerdata *rd, methoddesc *nmd)
{
	s4         *mcodeptr;               /* code generation pointer            */
	s4          nativeparams;
	s4          stackframesize;
	methoddesc *md;
	s4          i, j;
	s4          t;
	s4          a, s1, s2;

	/* initialize variables */
	md = m->parseddesc;
	nativeparams = (m->flags & ACC_STATIC) ? 2 : 1;

	/* calculate stackframe size */
	/* TODO: if you change this, don't forget to change REG_SP some lines below!!! */
	stackframesize = 
		1 +                                             /* return address */
		nmd->memuse;

	/* create method header */
	(void) dseg_addaddress(cd, m);                          /* MethodPointer  */
	(void) dseg_adds4(cd, stackframesize * 4);              /* FrameSize      */
	(void) dseg_adds4(cd, 0);                               /* IsSync         */
	(void) dseg_adds4(cd, 0);                               /* IsLeaf         */
	(void) dseg_adds4(cd, 0);                               /* IntSave        */
	(void) dseg_adds4(cd, 0);                               /* FltSave        */
	(void) dseg_addlinenumbertablesize(cd);
	(void) dseg_adds4(cd, 0);                               /* ExTableSize    */

	/* initialize mcode variables */
	mcodeptr = (s4*) cd->mcodebase;
	cd->mcodeend = (s4 *) (cd->mcodebase + cd->mcodesize);

	/* TODO: remove this, only for debugging! */
	/*M_MOV(2, REG_LR);
	M_MOV(1, 0);
	ICONST(0, 0xab);
	M_LONGBRANCH(asm_debug);*/

	/* generate stub code */
	/* TODO: have you changed stackframesize??? */
	/* TODO: nmd->memuse=(stackframesize-1) */
	/* TODO: don't forget ... there is and M_ADD_IMM at the end of this stub!!! */
	M_STMFD(1<<REG_LR, REG_SP);
	if (nmd->memuse) {
		M_SUB_IMM(REG_SP, REG_SP, nmd->memuse * 4);
	}

	/* if function is static, check for initialized */
	if ((m->flags & ACC_STATIC) && !m->class->initialized) {
		codegen_addpatchref(cd, mcodeptr, PATCHER_clinit, m->class);
		if (opt_showdisassemble) M_NOP;
	}

	/* call trace function */
	if (runverbose) {
		/* save integer and float argument registers */
		M_STMFD(0x0f | (1<<REG_LR) | (1<<REG_IP), REG_SP);

		/* call argument tracer */
		a = dseg_addaddress(cd, m);
		M_SUB_IMM(REG_SP, REG_SP, (2+2+1)*4);  /* get stack space for methodptr */
		ICONST(REG_ITMP1, 0);
		M_STR(2, REG_SP, 0);
		M_STR(REG_ITMP1, REG_SP, 4);
		M_STR(3, REG_SP, 8);
		M_STR(REG_ITMP1, REG_SP, 12);
		M_MOV(2, 1);
		M_MOV(1, REG_ITMP1);
		M_MOV(3, REG_ITMP1);
		M_LDR(REG_ITMP1, REG_IP, a);           /* store method pointer to stack */
		M_STR(REG_ITMP1, REG_SP, 16);

		M_LONGBRANCH(builtin_trace_args);      /* call the tracer (long branch) */

		M_ADD_IMM(REG_SP, REG_SP, (2+2+1)*4);  /* free stack space of methodptr */

		/* restore integer and float argument registers */
		M_LDMFD(0x0f | (1<<REG_LR) | (1<<REG_IP), REG_SP);
	}

	/* save integer and float argument registers */
	/*M_STMFD(0x0f | (1<<REG_IP), REG_SP);*/
	/* TODO: floating point */

	/* create native stack info */
	/*a = dseg_addaddress(cd, builtin_asm_get_stackframeinfo);*/
	/*M_ALD(REG_PV, REG_PV, off);*/
	/*M_MOV(REG_LR, REG_PC);
	M_LDR(REG_PC, REG_IP, a);*/
	/*disp = (s4) ((u1 *) mcodeptr - cd->mcodebase);
	M_LDA(REG_PV, REG_RA, -disp);*/

	/* restore integer and float argument registers */
	/*M_LDMFD(0x0f | (1<<REG_IP), REG_SP);*/

	/* copy or spill arguments to new locations */
	/* ATTENTION: the ARM has only integer argument registers! */
	for (i = md->paramcount - 1, j = i + nativeparams; i >= 0; i--, j--) {
		t = md->paramtypes[i].type;
		if (!md->params[i].inmemory) {
			s1 = ARGUMENT_REGS(t, md->params[i].regoff);
			if (!nmd->params[j].inmemory) {
				s2 = ARGUMENT_REGS(t, nmd->params[j].regoff);
				M_INTMOVE_TYPED(t, s1, s2);
			} else {
				s2 = nmd->params[j].regoff;
				if (IS_2_WORD_TYPE(t)) {
					M_STR(GET_LOW_REG(s1), REG_SP, (s2+0) * 4);
					M_STR(GET_HIGH_REG(s1), REG_SP, (s2+1) * 4);
				} else {
					M_STR(s1, REG_SP, s2 * 4);
				}
			}
		} else {
			/* TODO: there can be two word types on stack as well!!! */
			if (IS_2_WORD_TYPE(t)) {
				M_BREAKPOINT(0xabcd);
				dolog("NATIVESTUB: 2_WORD_TYPE on stack ... this is SEVERE (%s.%s)", m->class->name->text, m->name->text);
			}
			s1 = md->params[i].regoff + stackframesize;
			s2 = nmd->params[j].regoff;
			if (IS_2_WORD_TYPE(t)) {
				M_LDR(REG_ITMP1, REG_SP, (s1+0) * 4);
				M_LDR(REG_ITMP2, REG_SP, (s1+1) * 4);
				M_STR(REG_ITMP1, REG_SP, (s2+0) * 4);
				M_STR(REG_ITMP2, REG_SP, (s2+1) * 4);
			} else {
				M_LDR(REG_ITMP1, REG_SP, s1 * 4);
				M_STR(REG_ITMP1, REG_SP, s2 * 4);
			}
		}
	}

	/* put class into second argument register */
	if (m->flags & ACC_STATIC) {
		a = dseg_addaddress(cd, m->class);
		M_LDR(rd->argintregs[1], REG_IP, a);
	}

	/* put env into first argument register */
	a = dseg_addaddress(cd, &env);
	M_LDR(rd->argintregs[0], REG_IP, a);

	/* call debugger */
	if (0) {
		ASM_DEBUG_PREPARE;
		M_MOV(3,2);
		M_MOV(2,1);
		M_MOV(1,0);
		ICONST(0, 0xcc);
		ASM_DEBUG_EXECUTE;
	}

	/* do the native function call */
#if !defined(ENABLE_STATICVM)
	if (f == NULL) {
		codegen_addpatchref(cd, mcodeptr, PATCHER_resolve_native, m);
		if (opt_showdisassemble) M_NOP;
	}
#endif
	a = dseg_addaddress(cd, f);
	/*M_LDR(REG_IP, REG_IP, a);*/           /* load adress of native method       */
	M_MOV(REG_LR, REG_PC);              /* set return address                 */
	M_LDR(REG_PC, REG_IP, a);           /* call native method                 */

	/* recompute ip from pc */
	/* TODO: this is only needed because of the tracer ... do we really need it? */
	s1 = (s4) ((u1*) mcodeptr - cd->mcodebase);
	s1 += 8;                         /* we use PC relative addr.  */
	M_RECOMPUTE_IP(s1);

	/* call debugger */
	if (0) {
		ASM_DEBUG_PREPARE;
		ICONST(3,0);
		ICONST(2,0);
		M_MOV(1,0);
		ICONST(0, 0xdd);
		ASM_DEBUG_EXECUTE;
	}

	/* call trace function */
	if (runverbose) {
		M_STMFD(0x0f | (1<<REG_LR) | (1<<REG_IP), REG_SP);

		a = dseg_addaddress(cd, m);
		M_MOV(2, 1);
		M_MOV(1, 0);
		M_LDR(0, REG_IP, a);

		M_LONGBRANCH(builtin_displaymethodstop);

		M_LDMFD(0x0f | (1<<REG_LR) | (1<<REG_IP), REG_SP);
	}

	/* finish stub code, return to caller */
	if (nmd->memuse) {
		M_ADD_IMM(REG_SP, REG_SP, nmd->memuse * 4);
	}
	M_LDMFD(1<<REG_PC, REG_SP);
	/*M_MOV(REG_PC, REG_LR);*/              /* return to caller                   */

	/* process patcher calls */
	{
	patchref *pref;
	u4        mcode;
	s4       *tmpmcodeptr;
	s4       *xcodeptr;

	for (pref = cd->patchrefs; pref != NULL; pref = pref->next) {
		/* check code segment size */
		MCODECHECK(50); /* TODO */

		/* Get machine code which is patched back in later. The call is   */
		/* 1 instruction word long.                                       */
		xcodeptr = (s4 *) (cd->mcodebase + pref->branchpos);
		mcode = *xcodeptr;

		/* patch in the call to call the following code (done at compile  */
		/* time)                                                          */
		tmpmcodeptr = mcodeptr;         /* save current mcodeptr          */
		mcodeptr = xcodeptr;            /* set mcodeptr to patch position */
		M_BL(tmpmcodeptr - (xcodeptr + 1) - 1);
		mcodeptr = tmpmcodeptr;         /* restore the current mcodeptr   */

		/* create stack frame */
		M_SUB_IMM(REG_SP, REG_SP, 5*4);

		/* move return address onto stack */
		M_STR(REG_LR, REG_SP, 4*4);

		/* move pointer to java_objectheader onto stack */
#if defined(USE_THREADS) && defined(NATIVE_THREADS)
		log_text("!!! PATCHER: threads not implemented yet");
#else
		M_EOR(REG_ITMP3, REG_ITMP3, REG_ITMP3);
		M_STR(REG_ITMP3, REG_SP, 3*4);
#endif

		/* move machine code onto stack */
		a = dseg_adds4(cd, mcode);
		M_LDR(REG_ITMP3, REG_IP, a);
		M_STR(REG_ITMP3, REG_SP, 2*4);

		/* move class/method/field reference onto stack */
		a = dseg_addaddress(cd, pref->ref);
		M_LDR(REG_ITMP3, REG_IP, a);
		M_STR(REG_ITMP3, REG_SP, 1*4);

		/* move patcher function pointer onto stack */
		a = dseg_addaddress(cd, pref->patcher);
		M_LDR(REG_ITMP3, REG_IP, a);
		M_STR(REG_ITMP3, REG_SP, 0*4);

		/* finally call the patcher via asm_wrapper_patcher */
		/* ATTENTION: don't use REG_IP here, because some patchers need it */
		a = dseg_addaddress(cd, asm_wrapper_patcher);
		M_LDR(REG_PC, REG_IP, a);
	}
	}

	codegen_finish(m, cd, (s4) ((u1*) mcodeptr - cd->mcodebase));

	/* synchronize instruction cache */
	asm_sync_instruction_cache();

	return m->entrypoint;
}


/* asm_debug *******************************************************************

   Lazy debugger!

*******************************************************************************/

s4 asm_debug_indent = 0;
void asm_debug_intern(int a1, int a2, int a3, int a4)
{
	s4 i;
	if ((a1 == 0xaa) && (asm_debug_indent < 20))
		asm_debug_indent++;
	printf("===> ");
	for (i=0; i<asm_debug_indent; i++) printf("- ");
	if ((a1 == 0xbb) && (asm_debug_indent > 0))
		asm_debug_indent--;
	if (a1 == 0xab)
		printf("INVOKATION RESULT - ");
	if (a1 == 0xff)
		printf("INTERNAL EXCEPTION - ");
	if (a1 == 0xea)
		printf("ENTERING ASM_CALL_JAVAFUNCTION - ");
	if (a1 == 0xeb)
		printf("LEFT ASM_CALLJAVAFUNCTION - ");
	printf("got asm_debug(%p, %p, %p, %p)\n",(void*)a1,(void*)a2,(void*)a3,(void*)a4);
        if (a1 == 0x88) {
                printf("i want to write back method code ptr\n");
		disassinstr((void*)(a4-12));
		disassinstr((void*)(a4-8));
		disassinstr((void*)(a4-4));
		disassinstr((void*)a4);
		printf("\n");
        }
}
void asm_debug(int a1, int a2, int a3, int a4)
{
	printf("===> i am going to exit after this debugging message!\n");
	asm_debug_intern(a1,a2,a3,a4);
	throw_cacao_exception_exit(string_java_lang_InternalError, "leave you now");
}


/*
 * These are local overrides for various environment variables in Emacs.
 * Please do not remove this and leave it at the end of the file, where
 * Emacs will automagically detect them.
 * ---------------------------------------------------------------------
 * Local variables:
 * mode: c
 * indent-tabs-mode: t
 * c-basic-offset: 4
 * tab-width: 4
 * End:
 */
