/* vm/jit/arm/codegen.h - code generation macros and definitions for arm

   Copyright (C) 1996-2005 R. Grafl, A. Krall, C. Kruegel, C. Oates,
   R. Obermaisser, M. Platter, M. Probst, S. Ring, E. Steiner,
   C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich, J. Wenninger,
   Institut f. Computersprachen - TU Wien

   This file is part of CACAO.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   Contact: cacao@complang.tuwien.ac.at

   Authors: Michael Starzinger

   $Id: codegen.h,v 1.16 2005/07/12 16:07:06 michi Exp $

*/


#ifndef _CODEGEN_H
#define _CODEGEN_H


/* helper macros for generating code ******************************************/

#ifdef STATISTICS
#define COUNT_SPILLS count_spills++
#else
#define COUNT_SPILLS
#endif

/* load_var_to_reg_xxx:
   this function generates code to fetch data from a pseudo-register
   into a real register. 
   If the pseudo-register has actually been assigned to a real 
   register, no code will be emitted, since following operations
   can use this register directly.
   v:      pseudoregister to be fetched from
   tempnr: temporary register to be used if v is actually spilled to ram
   regnr:  the register number, where the operand can be found after 
           fetching (this wil be either tempregnum or the register
           number allready given to v)
*/
#define load_var_to_reg_int(regnr,v,tempnr) { \
	if ((v)->flags & INMEMORY) { \
		COUNT_SPILLS; \
		if (IS_2_WORD_TYPE((v)->type)) { \
			M_LDR(GET_LOW_REG((tempnr)), REG_SP, 4 * (v)->regoff); \
			M_LDR(GET_HIGH_REG((tempnr)), REG_SP, 4 * (v)->regoff + 4); \
		} else \
			M_LDR(tempnr, REG_SP, 4 * (v)->regoff); \
		regnr = tempnr; \
	} else regnr = (v)->regoff; \
}
#define load_var_to_reg_int_low(regnr,v,tempnr) { \
	if ((v)->flags & INMEMORY) { \
		COUNT_SPILLS; \
		M_LDR(tempnr, REG_SP, 4 * (v)->regoff); \
		regnr = tempnr; \
	} else regnr = GET_LOW_REG((v)->regoff); \
}
#define load_var_to_reg_int_high(regnr,v,tempnr) { \
	if ((v)->flags & INMEMORY) { \
		COUNT_SPILLS; \
		M_LDR(tempnr, REG_SP, 4 * (v)->regoff + 4); \
		regnr = tempnr; \
	} else regnr = GET_HIGH_REG((v)->regoff); \
}
#define load_var_to_reg_flt(regnr,v,tempnr) { \
	if ((v)->flags & INMEMORY) { \
		COUNT_SPILLS; \
		if ((v)->type == TYPE_DBL) \
			M_LDFD(tempnr, REG_SP, 4 * (v)->regoff); \
		else \
			M_LDFS(tempnr, REG_SP, 4 * (v)->regoff); \
		regnr = tempnr; \
	} else regnr = (v)->regoff; \
}

/* store_reg_to_var_xxx:
   This function generates the code to store the result of an operation
   back into a spilled pseudo-variable.
   If the pseudo-variable has not been spilled in the first place, this 
   function will generate nothing.
   v:      Pseudovariable
   tempnr: Number of the temporary registers as returned by
           reg_of_var.
*/
#define store_reg_to_var_int(v,tempnr) { \
	if ((v)->flags & INMEMORY) { \
		COUNT_SPILLS; \
		if (IS_2_WORD_TYPE((v)->type)) { \
			M_STR(GET_LOW_REG((tempnr)), REG_SP, 4 * (v)->regoff); \
			M_STR(GET_HIGH_REG((tempnr)), REG_SP, 4 * (v)->regoff + 4); \
		} else \
			M_STR(tempnr, REG_SP, 4 * (v)->regoff); \
	} \
}
#define store_reg_to_var_flt(v,tempnr) { \
	if ((v)->flags & INMEMORY) { \
		COUNT_SPILLS; \
		if ((v)->type == TYPE_DBL) \
			M_STFD(tempnr, REG_SP, 4 * (v)->regoff); \
		else \
			M_STFS(tempnr, REG_SP, 4 * (v)->regoff); \
	} \
}


#define MCODECHECK(icnt) \
	if ((mcodeptr + (icnt)) > cd->mcodeend) \
		mcodeptr = codegen_increase(cd, (u1 *) mcodeptr)

#define IS_IMM(val) \
	( ((val) >= 0) && ((val) <= 255) )
	/* TODO: correct this! */

#define  CHECK_OFFSET(off,max) \
	if ((off) > (max) || (off) < -(max)) printf("CHECK_OFFSET: offset out of range: %x (>%x) SEVERE ERROR!!!\n", ((off)<0)?-(off):off, max)


/* lazy debugger **************************************************************/

void asm_debug(int a1, int a2, int a3, int a4);
void asm_debug_intern(int a1, int a2, int a3, int a4);

/* if called with this macros, it can be placed nearly anywhere */
/* almost all registers are saved and restored afterwards       */
/* it uses a long branch to call the asm_debug_intern (no exit) */
#define ASM_DEBUG_PREPARE \
	M_STMFD(0x7fff, REG_SP)
#define ASM_DEBUG_EXECUTE \
	M_LONGBRANCH(asm_debug_intern); \
	M_LDMFD(0x7fff, REG_SP)


/* macros to create code ******************************************************/

/* the condition field */
#define COND_EQ 0x0  /* Equal        Z set   */
#define COND_NE 0x1  /* Not equal    Z clear */
#define COND_CS 0x2  /* Carry set    C set   */
#define COND_CC 0x3  /* Carry clear  C clear */
#define COND_MI 0x4  /* Negative     N set   */
#define COND_PL 0x5  /* Positive     N clear */
#define COND_VS 0x6  /* Overflow     V set   */
#define COND_VC 0x7  /* No overflow  V clear */
#define COND_HI 0x8  /* Unsigned higher      */
#define COND_LS 0x9  /* Unsigned lower, same */
#define COND_GE 0xA  /* Sig. greater, equal  */
#define COND_LT 0xB  /* Sig. less than       */
#define COND_GT 0xC  /* Sig. greater than    */
#define COND_LE 0xD  /* Sig. less, equal     */
#define COND_AL 0xE  /* Always               */
#define CONDNV  0xF  /* Special (see A3-5)   */
#define UNCOND COND_AL

/* data processing operation: M_DAT
   cond ... conditional execution
   op ..... opcode
   d ...... destination reg
   n ...... source reg
   S ...... update condition codes
   I ...... switch to immediate mode
   shift .. shifter operand
*/
#define M_DAT(cond,op,d,n,S,I,shift) \
	*(mcodeptr++) = (((cond) << 28) | ((op) << 21) | ((d) << 12) | ((n) << 16) | \
					((I) << 25) | ((S) << 20)| ((shift) & 0x00000fff))

/* load and store instruction: M_MEM
   cond ... conditional execution
   L ...... load (L=1) or store (L=0)
   B ...... unsigned byte (B=1) or word (B=0)
   d ...... destination reg
   n ...... base reg for addressing
   adr .... addressing mode specific
*/
#define M_MEM(cond,L,B,d,n,adr,I,P,U,W) \
	*(mcodeptr++) = (((cond) << 28) | (1 << 26) | ((L) << 20) | ((B) << 22) | ((d) << 12) | ((n) << 16) | \
					((adr) & 0x0fff) | ((I) << 25) | ((P) << 24) | ((U) << 23) | ((W) << 21))

/* load and store instruction: M_MEM2
   cond ... conditional execution
   L ...... load (L=1) or store (L=0)
   H ...... halfword (H=1) or signed byte (H=0)
   S ...... signed (S=1) or unsigned (S=0) halfword
   d ...... destination reg
   n ...... base reg for addressing
   adr .... addressing mode specific
*/
#define M_MEM2(cond,L,H,S,d,n,adr,I,P,U,W) \
	*(mcodeptr++) = (((cond) << 28) | (1 << 22) | (0x9 << 4) | ((L) << 20) | ((H) << 5) | ((S) << 6) | ((d) << 12) | ((n) << 16) | \
					((adr) & 0x0f) | (((adr) & 0xf0) << (8-4)) | ((I) << 22) | ((P) << 24) | ((U) << 23) | ((W) << 21))

/* load and store multiple instruction: M_MEM_MULTI
   cond ... conditional execution
   L ...... load (L=1) or store (L=0)
   S ...... special (see "The ARM ARM A3-21")
   regs ... register list
   n ...... base reg for addressing
*/
#define M_MEM_MULTI(cond,L,S,regs,n,P,U,W) \
	*(mcodeptr++) = (((cond) << 28) | (1 << 27) | ((L) << 20) | ((S) << 22) | ((n) << 16) | \
					((regs) & 0xffff) | ((P) << 24) | ((U) << 23) | ((W) << 21))

/* branch and branch with link: M_BRA
   cond ... conditional execution
   L ...... branch with link (L=1)
   offset . 24bit offset
  */
#define M_BRA(cond,L,offset) \
	*(mcodeptr++) = (((cond) << 28) | (0x5 << 25) | ((L) << 24) | ((offset) & 0x00ffffff))

/* multiplies: M_MULT
   cond ... conditional execution
   d ...... destination register
   n, m ... source registers
   S ...... update conditional codes
   A ...... accumulate flag (enables third source)
   s ...... third source register
*/
#define M_MULT(cond,d,n,m,S,A,s) \
	*(mcodeptr++) = (((cond) << 28) | ((d) << 16) | ((n) << 8) | (m) | (0x09 << 4) | ((S) << 20) | ((A) << 21) | ((s) << 12))

/* no operation (mov r0,r0): M_NOP */
#define M_NOP \
	*(mcodeptr++) = (0xe1a00000)

/* software breakpoint (only v5 and above): M_BREAKPOINT */
#define M_BREAKPOINT(imm) \
	*(mcodeptr++) =  (0x0e12 << 20) | (0x07 << 4) | (((imm) & 0xfff0) << (8-4)) | ((imm) & 0x0f)

/* Floating-Point Coprocessor Data Operations: M_CPDO
   cond ... conditional execution
   op ..... opcode
   D ...... dyadic (D=0) or monadic (D=1) instruction
   Fd ..... destination float-register
   Fn ..... source float-register
   Fm ..... source float-register or immediate
*/
#define M_CPDO(cond,op,D,Fd,Fn,Fm) \
	*(mcodeptr++) = (((cond) << 28) | (0x0e << 24) | (1 << 8) | ((op) << 20) | ((D) << 15) | ((Fd) << 12) | ((Fn) << 16) | ((Fm) & 0x0f))

/* Floating-Point Coprocessor Data Transfer: M_CPDT
   cond ... conditional execution
   L ...... load (L=1) or store (L=0)
   Fd ..... destination float-register
   n ...... base reg for addressing
*/
#define M_CPDT(cond,L,T1,T0,Fd,n,off,P,U,W) \
	*(mcodeptr++) = (((cond) << 28) | (0x0c << 24) | (1 << 8) | ((L) << 20) | ((T1) << 22) | ((T0) << 15) | ((Fd) << 12) | ((n) << 16) | \
					((off) & 0xff) | ((P) << 24) | ((U) << 23) | ((W) << 21))

/* Floating-Point Coprocessor Register Transfer: M_CPRT
*/
#define M_CPRT(cond,L,d,Fn,Fm) \
	*(mcodeptr++) = (((cond) << 28) | (0x0e << 24) | (1 << 8) | (1 << 4) | ((L) << 20) | ((d) << 12) | ((Fn) << 16) | (Fm))

/* used to store values! */
#define DCD(val) \
	*(mcodeptr++) = val

/* used to directly access shifter; insert this as shifter operand! */
#define REG_LSL(reg, shift) ( (((shift) & 0x1f) << 7) | ((reg) & 0x0f) )
#define REG_LSR(reg, shift) ( (((shift) & 0x1f) << 7) | ((reg) & 0x0f) | (1 << 5) )
#define REG_ASR(reg, shift) ( (((shift) & 0x1f) << 7) | ((reg) & 0x0f) | (1 << 6) )
#define REG_LSL_REG(reg, s) ( (((s) & 0x0f) << 8) | ((reg) & 0x0f) | (1 << 4) )

/* macros for all arm instructions ********************************************/

#define M_ADD(d,a,b)       M_DAT(UNCOND,0x04,d,a,0,0,b)         /* d = a +  b */
#define M_ADC(d,a,b)       M_DAT(UNCOND,0x05,d,a,0,0,b)         /* d = a +  b (with Carry) */
#define M_SUB(d,a,b)       M_DAT(UNCOND,0x02,d,a,0,0,b)         /* d = a -  b */
#define M_SBC(d,a,b)       M_DAT(UNCOND,0x06,d,a,0,0,b)         /* d = a -  b (with Carry) */
#define M_AND(d,a,b)       M_DAT(UNCOND,0x00,d,a,0,0,b)         /* d = a &  b */
#define M_ORR(d,a,b)       M_DAT(UNCOND,0x0c,d,a,0,0,b)         /* d = a |  b */
#define M_EOR(d,a,b)       M_DAT(UNCOND,0x01,d,a,0,0,b)         /* d = a ^  b */
#define M_TST(a,b)         M_DAT(UNCOND,0x08,0,a,1,0,b)         /* TST a &  b */
#define M_TEQ(a,b)         M_DAT(UNCOND,0x09,0,a,1,0,b)         /* TST a ^  b */
#define M_CMP(a,b)         M_DAT(UNCOND,0x0a,0,a,1,0,b)         /* TST a -  b */
#define M_MOV(d,b)         M_DAT(UNCOND,0x0d,d,0,0,0,b)         /* d =      b */
#define M_ADD_S(d,a,b)     M_DAT(UNCOND,0x04,d,a,1,0,b)         /* d = a +  b (update Flags) */
#define M_SUB_S(d,a,b)     M_DAT(UNCOND,0x02,d,a,1,0,b)         /* d = a -  b (update Flags) */

#define M_NEG(d,a)         M_DAT(UNCOND,0x03,d,a,0,1,0)         /* d = -a + 0 */

#define M_ADD_IMM(d,a,i)   M_DAT(UNCOND,0x04,d,a,0,1,i)         /* d = a +  i */
#define M_ADC_IMM(d,a,i)   M_DAT(UNCOND,0x05,d,a,0,1,i)         /* d = a +  i (with Carry) */
#define M_SUB_IMM(d,a,i)   M_DAT(UNCOND,0x02,d,a,0,1,i)         /* d = a -  i */
#define M_SBC_IMM(d,a,i)   M_DAT(UNCOND,0x06,d,a,0,1,i)         /* d = a -  i (with Carry) */
#define M_RSC_IMM(d,a,i)   M_DAT(UNCOND,0x07,d,a,0,1,i)         /* d = -a + i (with Carry) */
#define M_TST_IMM(a,i)     M_DAT(UNCOND,0x08,0,a,1,1,i)         /* TST a &  i */
#define M_TEQ_IMM(a,i)     M_DAT(UNCOND,0x09,0,a,1,1,i)         /* TST a ^  i */
#define M_CMP_IMM(a,i)     M_DAT(UNCOND,0x0a,0,a,1,1,i)         /* TST a -  i */
#define M_CMN_IMM(a,i)     M_DAT(UNCOND,0x0b,0,a,1,1,i)         /* TST a +  i */
#define M_MOV_IMM(d,i)     M_DAT(UNCOND,0x0d,d,0,0,1,i)         /* d =      i */
#define M_ADD_IMMS(d,a,i)  M_DAT(UNCOND,0x04,d,a,1,1,i)         /* d = a +  i (update Flags) */
#define M_SUB_IMMS(d,a,i)  M_DAT(UNCOND,0x02,d,a,1,1,i)         /* d = a -  i (update Flags) */
#define M_RSB_IMMS(d,a,i)  M_DAT(UNCOND,0x03,d,a,1,1,i)         /* d = -a + i (update Flags) */

#define M_ADDSUB_IMM(d,a,i) if((i)>=0) M_ADD_IMM(d,a,i); else M_SUB_IMM(d,a,-(i))
#define M_ADDLT_IMM(d,a,i) M_DAT(COND_LT,0x04,d,a,0,1,i)
#define M_ADDGT_IMM(d,a,i) M_DAT(COND_GT,0x04,d,a,0,1,i)
#define M_SUBLT_IMM(d,a,i) M_DAT(COND_LT,0x02,d,a,0,1,i)
#define M_SUBGT_IMM(d,a,i) M_DAT(COND_GT,0x02,d,a,0,1,i)

#define M_MUL(d,a,b)       M_MULT(UNCOND,d,a,b,0,0,0x0)         /* d = a *  b */


#define M_LDR(d,base,off)   CHECK_OFFSET(off,0x0fff); M_MEM(UNCOND,1,0,d,base,((off)<0)?-(off):off,0,1,((off)<0)?0:1,0)
#define M_STR(d,base,off)   CHECK_OFFSET(off,0x0fff); M_MEM(UNCOND,0,0,d,base,((off)<0)?-(off):off,0,1,((off)<0)?0:1,0)
/* TODO: these have to be fixed! */
#define M_LDRH(d,base,off)  CHECK_OFFSET(off,0x00ff); M_MEM2(UNCOND,1,1,0,d,base,off,1,1,1,0)
#define M_LDRSB(d,base,off) CHECK_OFFSET(off,0x00ff); M_MEM2(UNCOND,1,0,1,d,base,off,1,1,1,0)
#define M_STRH(d,base,off)  CHECK_OFFSET(off,0x00ff); M_MEM2(UNCOND,0,1,0,d,base,off,1,1,1,0)
#define M_STRB(d,base,off)  CHECK_OFFSET(off,0x0fff); M_MEM(UNCOND,0,1,d,base,off,0,1,1,0)

#define M_LDMFD(regs,base) M_MEM_MULTI(UNCOND,1,0,regs,base,0,1,1)
#define M_STMFD(regs,base) M_MEM_MULTI(UNCOND,0,0,regs,base,1,0,1)

#define M_LDFS(d,base,off) M_CPDT(UNCOND,1,0,0,d,base,((off)<0)?-(off)>>2:(off)>>2,1,((off)<0)?0:1,0)
#define M_LDFD(d,base,off) M_CPDT(UNCOND,1,0,1,d,base,((off)<0)?-(off)>>2:(off)>>2,1,((off)<0)?0:1,0)
#define M_STFS(d,base,off) M_CPDT(UNCOND,0,0,0,d,base,((off)<0)?-(off)>>2:(off)>>2,1,((off)<0)?0:1,0)
#define M_STFD(d,base,off) M_CPDT(UNCOND,0,0,1,d,base,((off)<0)?-(off)>>2:(off)>>2,1,((off)<0)?0:1,0)


#define M_ADFS(d,a,b)      M_CPDO(UNCOND,0x00,0,d,a,b)          /* d = a +  b */
#define M_SUFS(d,a,b)      M_CPDO(UNCOND,0x02,0,d,a,b)          /* d = a -  b */
#define M_MUFS(d,a,b)      M_CPDO(UNCOND,0x01,0,d,a,b)          /* d = a *  b */
#define M_DVFS(d,a,b)      M_CPDO(UNCOND,0x04,0,d,a,b)          /* d = a /  b */
#define M_RMFS(d,a,b)      M_CPDO(UNCOND,0x08,0,d,a,b)          /* d = a %  b */
#define M_ADFD(d,a,b)      M_CPDO(UNCOND,0x00,0,d,a,b) | (1<<7) /* d = a +  b */
#define M_SUFD(d,a,b)      M_CPDO(UNCOND,0x02,0,d,a,b) | (1<<7) /* d = a -  b */
#define M_MUFD(d,a,b)      M_CPDO(UNCOND,0x01,0,d,a,b) | (1<<7) /* d = a *  b */
#define M_DVFD(d,a,b)      M_CPDO(UNCOND,0x04,0,d,a,b) | (1<<7) /* d = a /  b */
#define M_RMFD(d,a,b)      M_CPDO(UNCOND,0x08,0,d,a,b) | (1<<7) /* d = a %  b */
#define M_MVFS(d,a)        M_CPDO(UNCOND,0x00,1,d,0,a)          /* d =      a */
#define M_MVFD(d,a)        M_CPDO(UNCOND,0x00,1,d,0,a) | (1<<7) /* d =      a */
#define M_CMF(a,b)         M_CPRT(UNCOND,1,0x0f,a,b) | (1<<23)  /* COMPARE a;  b */
#define M_FLTS(d,a)        M_CPRT(UNCOND,0,a,d,0)               /* d = (float) a */
#define M_FLTD(d,a)        M_CPRT(UNCOND,0,a,d,0) | (1<<7)      /* d = (float) a */
#define M_FIX(d,a)         M_CPRT(UNCOND,1,d,0,a)               /* d = (int)   a */


#define M_B(off)           M_BRA(UNCOND,0,off)    /* unconditional branch */
#define M_BL(off)          M_BRA(UNCOND,1,off)    /* branch and link      */
#define M_BEQ(off)         M_BRA(COND_EQ,0,off)   /* conditional branches */
#define M_BNE(off)         M_BRA(COND_NE,0,off)
#define M_BGE(off)         M_BRA(COND_GE,0,off)
#define M_BGT(off)         M_BRA(COND_GT,0,off)
#define M_BLT(off)         M_BRA(COND_LT,0,off)
#define M_BLE(off)         M_BRA(COND_LE,0,off)
#define M_BHI(off)         M_BRA(COND_HI,0,off)   /* unsigned conditional */


/* ICONST/LCONST:
   loads the integer/long value const into register d.
*/
#define ICONST(d,const) \
	if (IS_IMM(const)) { \
		M_MOV_IMM(d,const); \
	} else { \
		a = dseg_adds4(cd, (const)); \
		M_LDR((d), REG_IP, a); \
	}
#define LCONST(d,const) \
	ICONST(GET_HIGH_REG((d)), (s4) ((s8) (const) >> 32)); \
	ICONST(GET_LOW_REG((d)), (s4) ((s8) (const)));
#define ICONST_DIRECT(d,const) \
	if ((const) > 0x0ffff) printf("ICONST_DIRECT: integer constant too high: %x (>0xffff) SEVERE ERROR!!!\n", const); \
	M_MOV_IMM(d,(const) & 0xff); \
        M_ADD_IMM(d,d,(((const) >>  8) & 0xff) | (((12) & 0x0f) << 8)); /*\
       	M_ADD_IMM(d,d,(((const) >> 16) & 0xff) | ((( 8) & 0x0f) << 8)); \
        M_ADD_IMM(d,d,(((const) >> 24) & 0xff) | ((( 4) & 0x0f) << 8)); \ */

/* M_RECOMPUTE_IP:
   used to recompute our IP (something like PV) out of the current PC
   ATTENTION: if you change this, you habe to look at asm_call_jit_compiler!
*/
#define M_RECOMPUTE_IP(disp) \
        if (IS_IMM(disp)) { \
                M_SUB_IMM(REG_IP, REG_PC, disp); \
        } else { \
		M_MOV(REG_IP, REG_PC); \
                ICONST_DIRECT(REG_ITMP3, disp); \
                M_SUB(REG_IP, REG_IP, REG_ITMP3); \
        }

/* M_INTMOVE:
   generates an integer-move from register a to b.
   if a and b are the same int-register, no code will be generated.
*/
#define M_INTMOVE(a,b) if ((a) != (b)) { M_MOV(b, a); }
#define M_INTMOVE_TYPED(t,a,b) \
	if (IS_2_WORD_TYPE(t)) { \
		/* TODO: check this! (what if [R1/R2]=[R0/R1] or something like this???)*/ \
		M_INTMOVE(GET_HIGH_REG((a)), GET_HIGH_REG((b))); \
		M_INTMOVE(GET_LOW_REG((a)), GET_LOW_REG((b))); \
	} else { \
		M_INTMOVE((a), (b)); \
	}

/* M_FLTMOVE:
   generates a floating-point-move from register a to b.
   if a and b are the same float-register, no code will be generated.
*/
#define M_FLTMOVE_TYPED(t,Fa,Fb) \
	if ((Fa) != (Fb)) { \
		if ((t) == TYPE_DBL) \
			M_MVFD(Fb, Fa); \
		else \
			M_MVFS(Fb, Fa); \
	}

/* M_FLTLOAD:
   TODO: improve this! (this is really ugly!!!)
   loads the value of the integer-register a (argument or result) into
   float-register Fb.
*/
#define M_FLTLOAD_TYPED(t,a,Fb) \
	if ((t) == TYPE_FLT) { \
		M_SUB_IMM(REG_SP, REG_SP, 4); \
		M_STR(a, REG_SP, 0); \
		M_LDFS(Fb, REG_SP, 0); \
		M_ADD_IMM(REG_SP, REG_SP, 4); \
	} else { \
		M_SUB_IMM(REG_SP, REG_SP, 8); \
		M_STR(GET_LOW_REG(a), REG_SP, 0); \
		M_STR(GET_HIGH_REG(a), REG_SP, 4); \
		M_LDFD(Fb, REG_SP, 0); \
		M_ADD_IMM(REG_SP, REG_SP, 8); \
	}

#define M_FLTSTORE_TYPED(t,Fa,b) \
	if ((t) == TYPE_FLT) { \
		M_SUB_IMM(REG_SP, REG_SP, 4); \
		M_STFS(Fa, REG_SP, 0); \
		M_LDR(b, REG_SP, 0); \
		M_ADD_IMM(REG_SP, REG_SP, 4); \
	} else { \
		M_SUB_IMM(REG_SP, REG_SP, 8); \
		M_STFD(Fa, REG_SP, 0); \
		M_LDR(GET_LOW_REG(b), REG_SP, 0); \
		M_LDR(GET_HIGH_REG(b), REG_SP, 4); \
		M_ADD_IMM(REG_SP, REG_SP, 8); \
	}

/* M_COPY:
   TODO: document me
*/
#define M_COPY(from,to) \
	d = reg_of_var(rd, to, REG_ITMP1); \
	if ((from->regoff != to->regoff) || \
	    ((from->flags ^ to->flags) & INMEMORY)) { \
		if (IS_FLT_DBL_TYPE(from->type)) { \
			load_var_to_reg_flt(s1, from, d); \
			M_FLTMOVE_TYPED(from->type,s1,d); \
			store_reg_to_var_flt(to, d); \
		} else { \
			load_var_to_reg_int(s1, from, d); \
			M_INTMOVE_TYPED(from->type,s1,d); \
			store_reg_to_var_int(to, d); \
		} \
	}

/* M_COMPARE:
   generates the compare part of an if-sequece
   uses M_CMP or M_CMP_IMM to do the compare
   ATTENTION: uses REG_ITMP2 as intermediate register
*/
#define M_COMPARE(reg, val, cond) \
	if (IS_IMM(val)) { \
		/* M_CMP_IMM */ M_DAT(cond,0x0a,0,(reg),1,1,(val)); \
	} else if(IS_IMM(-(val))) { \
		/* M_CMN_IMM */ M_DAT(cond,0x0b,0,(reg),1,1,-(val)); \
	} else { \
		ICONST(REG_ITMP2, (val)); \
		/* M_CMP */ M_DAT(cond,0x0a,0,(reg),1,0,REG_ITMP2); \
	}

/* M_LONGBRANCH:
   performs a long branch to an absolute address with return address in LR
   takes up 3 bytes of code space; address is hard-coded into code
*/
#define M_LONGBRANCH(adr) \
	M_ADD_IMM(REG_LR, REG_PC, 4); \
	M_LDR(REG_PC, REG_PC, -4); \
	DCD((s4) adr);


/* function gen_resolvebranch **************************************************
   ip ... pointer to instruction after branch (void*)
   so ... offset of instruction after branch  (s4)
   to ... offset of branch target             (s4) */
#define gen_resolvebranch(ip,so,to) \
	((s4*) (ip))[-1] |= ((s4) (to) - (so) - 1) >> 2 & 0x0ffffff

/* function gen_nullptr_check *************************************************/
#define gen_nullptr_check(objreg) \
	if (checknull) { \
		M_TST((objreg), (objreg)); \
		M_BEQ(-2); \
		codegen_addxnullrefs(cd, mcodeptr); \
	}

/* function gen_bound_check ***************************************************
   ATTENTION: uses REG_ITMP3 as intermediate register */
/* TODO: maybe use another reg instead of REG_ITMP3! */
#define gen_bound_check(arrayref,index) \
	if (checkbounds) { \
		M_LDR(REG_ITMP3, (arrayref), OFFSET(java_arrayheader, size)); \
		M_CMP((index), REG_ITMP3); \
		M_BGE(-2); \
		codegen_addxboundrefs(cd, mcodeptr, s2); \
	}


#endif /* _CODEGEN_H */


/*
 * These are local overrides for various environment variables in Emacs.
 * Please do not remove this and leave it at the end of the file, where
 * Emacs will automagically detect them.
 * ---------------------------------------------------------------------
 * Local variables:
 * mode: c
 * indent-tabs-mode: t
 * c-basic-offset: 4
 * tab-width: 4
 * End:
 */
